# isort: off
# TODO(pygeos): remove this once shapely is the default backend for geopandas.
# Force raster_tools._compat to be loaded before geopandas when running tests
import raster_tools  # noqa: F401

# isort: on

import geopandas as gpd
import numpy as np
import pytest
import xarray as xr
from shapely.geometry import LinearRing, LineString, Point, Polygon, box

from raster_tools import line_stats
from raster_tools.raster import Raster
from tests.utils import assert_rasters_similar, assert_valid_raster


@pytest.mark.parametrize(
    "geoms,n",
    [
        (gpd.GeoSeries(), 0),
        (gpd.GeoSeries([Point(0, 0)]), 1),
        (gpd.GeoSeries([LineString([(0, 0), (1, 1), (2, 2)])]), 3),
        (gpd.GeoSeries([box(0, 0, 1, 1)]), 5),
        (gpd.GeoSeries([LinearRing([(0, 0), (1, 1), (1, 0)])]), 4),
        (
            gpd.GeoSeries(
                [
                    Polygon([(0, 0), (0, 2), (2, 2), (2, 0), (0, 0)]),
                    # Polygon with interior ring
                    Polygon(
                        [(0, 0), (0, 2), (2, 2), (2, 0), (0, 0)],
                        [[(1, 0), (1.5, 0.5), (1, 1), (0.5, 0.5), (1, 0)]],
                    ),
                ]
            ),
            15,
        ),
        (
            gpd.GeoSeries(
                [
                    Point(0, 0),
                    LineString([(0, 0), (1, 1), (2, 2)]),
                    Polygon([(0, 0), (0, 2), (2, 2), (2, 0), (0, 0)]),
                    # Polygon with interior ring
                    Polygon(
                        [(0, 0), (0, 2), (2, 2), (2, 0), (0, 0)],
                        [[(1, 0), (1.5, 0.5), (1, 1), (0.5, 0.5), (1, 0)]],
                    ),
                ]
            ),
            19,
        ),
    ],
)
def test__calculate_number_vertices(geoms, n):
    count = line_stats._calculate_number_vertices(geoms)
    assert count == n


template_22 = xr.DataArray(
    np.ones((2, 2)), dims=("y", "x"), coords=([1, 0], [0, 1])
).rio.write_crs("5070")
template_44 = xr.DataArray(
    np.ones((4, 4)),
    dims=("y", "x"),
    coords=(np.arange(0, 4)[::-1], np.arange(0, 4)),
).rio.write_crs("5070")


def like_template(template, data):
    out = template.copy()
    out.data = data
    return out


@pytest.mark.parametrize(
    "geoms,like,radius,truth,weight",
    [
        (
            gpd.GeoSeries([LineString([(0, 1), (1, 1)])], crs="EPSG:5070"),
            Raster(template_22),
            1,
            Raster(like_template(template_22, np.array([[1, 1], [0, 0]]))),
            False,
        ),
        (
            gpd.GeoSeries([LineString([(0, 1), (1, 1)])], crs="EPSG:5070"),
            Raster(template_22).chunk((1, 1, 1)),
            1,
            Raster(like_template(template_22, np.array([[1, 1], [0, 0]]))),
            False,
        ),
        (
            gpd.GeoSeries([LineString([(0, 1), (1, 1)])], crs="EPSG:5070"),
            Raster(template_22),
            2,
            Raster(like_template(template_22, np.array([[1, 1], [1, 1]]))),
            False,
        ),
        (
            gpd.GeoSeries(
                [LineString([(-0.5, 1), (1.5, 1)])], crs="EPSG:5070"
            ),
            Raster(template_22),
            1,
            Raster(like_template(template_22, np.array([[1.5, 1.5], [0, 0]]))),
            False,
        ),
        (
            gpd.GeoSeries(
                [LineString([(0, 3), (2, 3), (2, 0)])], crs="EPSG:5070"
            ),
            Raster(template_44),
            1,
            Raster(
                like_template(
                    template_44,
                    np.array(
                        [
                            [1, 2, 2, 0],
                            [0, 0, 2, 0],
                            [0, 0, 2, 0],
                            [0, 0, 1, 0],
                        ]
                    ),
                )
            ),
            False,
        ),
        (
            gpd.GeoSeries(
                [LineString([(0, 3), (2, 3), (2, 0)])], crs="EPSG:5070"
            ),
            Raster(template_44).chunk((1, 2, 2)),
            1,
            Raster(
                like_template(
                    template_44,
                    np.array(
                        [
                            [1, 2, 2, 0],
                            [0, 0, 2, 0],
                            [0, 0, 2, 0],
                            [0, 0, 1, 0],
                        ]
                    ),
                )
            ),
            False,
        ),
        (
            gpd.GeoDataFrame(
                {
                    "weights": [3, 4],
                    "geometry": [
                        LineString([(0, 3), (2, 3)]),
                        LineString([(2, 3), (2, 0)]),
                    ],
                },
                crs="EPSG:5070",
            ),
            Raster(template_44),
            1,
            Raster(
                like_template(
                    template_44,
                    np.array(
                        [
                            [1, 2, 2, 0],
                            [0, 0, 2, 0],
                            [0, 0, 2, 0],
                            [0, 0, 1, 0],
                        ]
                    ),
                )
            ),
            False,
        ),
        (
            gpd.GeoDataFrame(
                {
                    "weights": [3, 4],
                    "geometry": [
                        LineString([(0, 3), (2, 3)]),
                        LineString([(2, 3), (2, 0)]),
                    ],
                },
                crs="EPSG:5070",
            ),
            Raster(template_44),
            1,
            Raster(
                like_template(
                    template_44,
                    np.array(
                        [
                            [3, 6, 7, 0],
                            [0, 0, 8, 0],
                            [0, 0, 8, 0],
                            [0, 0, 4, 0],
                        ]
                    ),
                )
            ),
            True,
        ),
        (
            gpd.GeoDataFrame(
                {
                    "weights": [3, 4],
                    "geometry": [
                        LineString([(0, 3), (2, 3)]),
                        LineString([(2, 3), (2, 0)]),
                    ],
                },
                crs="EPSG:5070",
            ),
            Raster(template_44).chunk((1, 2, 2)),
            1,
            Raster(
                like_template(
                    template_44,
                    np.array(
                        [
                            [3, 6, 7, 0],
                            [0, 0, 8, 0],
                            [0, 0, 8, 0],
                            [0, 0, 4, 0],
                        ]
                    ),
                )
            ),
            True,
        ),
    ],
)
def test_length(geoms, like, radius, truth, weight):
    result = line_stats.length(
        geoms, like, radius, weighting_field="weights" if weight else None
    )

    assert_valid_raster(result)
    assert np.allclose(result, truth)
    assert_rasters_similar(result, like)
    assert result.mask.compute().sum() == 0



import numpy as np
import pytest

from raster_tools import focal, surface
from raster_tools.dtypes import I32
from raster_tools.general import band_concat
from raster_tools.masking import get_default_null_value
from raster_tools.raster import Raster
from tests.utils import assert_rasters_similar, assert_valid_raster


@pytest.fixture
def dem():
    return Raster("tests/data/raster/elevation.tif")


# TODO: add test for surface_area_3d


@pytest.mark.parametrize(
    "degrees,truth",
    [
        (None, Raster("tests/data/raster/slope.tif")),
        (True, Raster("tests/data/raster/slope.tif")),
        (False, Raster("tests/data/raster/slope_percent.tif")),
    ],
)
def test_slope(dem, degrees, truth):
    if degrees is None:
        slope = surface.slope(dem)
    else:
        slope = surface.slope(dem, degrees=degrees)
    assert_valid_raster(slope)
    assert_rasters_similar(slope, dem)
    assert_rasters_similar(slope, truth)
    assert slope._masked
    assert slope.null_value == get_default_null_value(slope.dtype)
    assert np.allclose(slope, truth.set_null_value(slope.null_value))
    assert slope.dtype == np.dtype("float64")


def test_aspect(dem):
    aspect = surface.aspect(dem)
    truth = Raster("tests/data/raster/aspect.tif")

    assert_valid_raster(aspect)
    assert_rasters_similar(aspect, dem)
    assert_rasters_similar(aspect, truth)
    assert aspect._masked
    assert aspect.null_value == get_default_null_value(aspect.dtype)
    assert np.allclose(aspect, truth.set_null_value(aspect.null_value), 4e-5)
    assert aspect.dtype == np.dtype("float64")


def test_curvature(dem):
    curv = surface.curvature(dem)
    truth = Raster("tests/data/raster/curv.tif")

    assert_valid_raster(curv)
    assert_rasters_similar(curv, dem)
    assert_rasters_similar(curv, truth)
    assert curv._masked
    assert curv.null_value == get_default_null_value(curv.dtype)
    # ESRI treats the edges as valid even though they are not.
    # surface.curvature does not so we ignore the edges in the comparison.
    mask = truth._ds.mask
    mask[:, 0, :] = True
    mask[:, -1, :] = True
    mask[:, :, 0] = True
    mask[:, :, -1] = True
    truth._ds["mask"] = mask
    assert np.allclose(curv, truth.set_null_value(curv.null_value))
    assert curv.dtype == np.dtype("float64")


def test_northing(dem):
    northing = surface.northing(dem)
    truth = Raster("tests/data/raster/northing.tif")

    assert_valid_raster(northing)
    assert_rasters_similar(northing, dem)
    assert_rasters_similar(northing, truth)
    assert dem._masked
    assert northing._masked
    assert northing.null_value == get_default_null_value(northing.dtype)
    assert np.allclose(
        northing, truth.set_null_value(northing.null_value), 1e-6, 1e-6
    )
    assert northing.dtype.kind == "f"

    northing = surface.northing(surface.aspect(dem), is_aspect=True)
    assert_valid_raster(northing)
    assert_rasters_similar(northing, dem)
    assert_rasters_similar(northing, truth)
    assert northing._masked
    assert northing.null_value == get_default_null_value(northing.dtype)
    assert np.allclose(
        northing, truth.set_null_value(northing.null_value), 1e-6, 1e-6
    )
    assert northing.dtype.kind == "f"


def test_easting(dem):
    easting = surface.easting(dem)
    truth = Raster("tests/data/raster/easting.tif")

    assert_valid_raster(easting)
    assert_rasters_similar(easting, dem)
    assert_rasters_similar(easting, truth)
    assert dem._masked
    assert easting._masked
    assert easting.null_value == get_default_null_value(easting.dtype)
    assert np.allclose(
        easting, truth.set_null_value(easting.null_value), 1e-6, 1e-6
    )
    assert easting.dtype.kind == "f"

    easting = surface.easting(surface.aspect(dem), is_aspect=True)
    assert_valid_raster(easting)
    assert_rasters_similar(easting, dem)
    assert_rasters_similar(easting, truth)
    assert easting._masked
    assert easting.null_value == get_default_null_value(easting.dtype)
    assert np.allclose(
        easting, truth.set_null_value(easting.null_value), 1e-6, 1e-6
    )
    assert easting.dtype.kind == "f"


def test_hillshade(dem):
    hill = surface.hillshade(dem)
    truth = Raster("tests/data/raster/hillshade.tif")

    assert_valid_raster(hill)
    assert_rasters_similar(hill, dem)
    assert_rasters_similar(hill, truth)
    assert hill._masked
    assert hill.null_value == 255
    assert np.allclose(hill, truth)
    assert hill.dtype == np.dtype("uint8")


def test_tpi():
    dem = Raster("tests/data/raster/elevation_small.tif")
    truth = ((dem - focal.focal(dem, "mean", (5, 11))) + 0.5).astype(
        I32, False
    )
    tpi = surface.tpi(dem, 5, 11)

    assert tpi.dtype == I32
    assert np.allclose(truth, tpi)
    assert tpi.null_value == get_default_null_value(I32)

    # Make sure that it works with multiple bands
    dem2 = dem + 100
    truth2 = ((dem2 - focal.focal(dem2, "mean", (5, 11))) + 0.5).astype(
        I32, False
    )
    truth = band_concat((truth, truth2))
    dem = band_concat((dem, dem2))

    tpi = surface.tpi(dem, 5, 11)

    assert np.allclose(truth, tpi)



import unittest

import dask
import dask_geopandas as dgpd
import geopandas as gpd
import numpy as np
import pytest
import rasterio as rio

from raster_tools import Raster
from raster_tools.dtypes import I32, U16
from raster_tools.masking import get_default_null_value
from raster_tools.vector import Vector, open_vectors
from tests.utils import assert_valid_raster


class TestOpenVectors(unittest.TestCase):
    def test_open_vectors(self):
        vs = open_vectors("tests/data/vector/Zones.gdb")

        self.assertIsInstance(vs, list)
        self.assertTrue(len(vs) == 2)
        self.assertIsInstance(vs[0], Vector)
        self.assertIsInstance(vs[1], Vector)
        self.assertTrue(len(vs[0]) == 10642)
        self.assertTrue(len(vs[1]) == 184)

        pods1 = open_vectors("tests/data/vector/Zones.gdb", layers="PODs")
        self.assertIsInstance(pods1, Vector)
        pods2 = open_vectors("tests/data/vector/Zones.gdb", layers=1)
        pods3 = open_vectors("tests/data/vector/Zones.gdb", layers=["PODs"])
        pods4 = open_vectors("tests/data/vector/Zones.gdb", layers=[1])
        for v in [pods2, pods3, pods4]:
            self.assertTrue(pods1.data.compute().equals(v.data.compute()))

    def test_open_vectors_errors(self):
        with self.assertRaises(ValueError):
            open_vectors("tests/data/vector/Zones.gdb", layers="dummy")
        with self.assertRaises(ValueError):
            open_vectors("tests/data/vector/Zones.gdb", layers=2)
        with self.assertRaises(ValueError):
            open_vectors("tests/data/vector/Zones.gdb", layers=-1)
        with self.assertRaises(ValueError):
            open_vectors("tests/data/vector/Zones.gdb", layers=[-1])

        with self.assertRaises(TypeError):
            open_vectors("tests/data/vector/Zones.gdb", layers=[0, "PODs"])
        with self.assertRaises(TypeError):
            open_vectors("tests/data/vector/Zones.gdb", layers={})


class TestVectorProperties(unittest.TestCase):
    def setUp(self):
        self.v = open_vectors("tests/data/vector/pods.shp")

    def test_table(self):
        self.assertTrue(hasattr(self.v, "table"))
        self.assertTrue(len(self.v) == len(self.v.table))
        # Table doesn't contain geometry. It is only vector attributes
        self.assertTrue("geometry" not in self.v.table.columns)
        self.assertTrue(
            self.v.table.columns.size == (self.v.data.columns.size - 1)
        )

    def test_size(self):
        self.assertTrue(hasattr(self.v, "size"))
        self.assertTrue(self.v.size == 184)

    def test_shape(self):
        self.assertTrue(hasattr(self.v, "shape"))
        self.assertIsInstance(self.v.shape, tuple)
        self.assertTrue(self.v.shape == dask.compute(self.v.table.shape)[0])

    def test_crs(self):
        self.assertTrue(hasattr(self.v, "crs"))
        self.assertIsInstance(self.v.crs, rio.crs.CRS)
        self.assertTrue(self.v.crs == rio.crs.CRS.from_epsg(5070))

    def test_field_schema(self):
        self.assertTrue(hasattr(self.v, "field_schema"))
        self.assertIsInstance(self.v.field_schema, dict)
        self.assertTrue(self.v.field_schema == self.v.table.dtypes.to_dict())

    def test_field_names(self):
        self.assertTrue(hasattr(self.v, "field_names"))
        self.assertIsInstance(self.v.field_names, list)
        self.assertTrue(self.v.field_names == self.v.table.columns.to_list())

    def test_field_dtypes(self):
        self.assertTrue(hasattr(self.v, "field_dtypes"))
        self.assertIsInstance(self.v.field_dtypes, list)
        self.assertTrue(
            self.v.field_dtypes == list(self.v.table.dtypes.to_dict().values())
        )

    def test_geometry(self):
        self.assertTrue(hasattr(self.v, "geometry"))
        self.assertIsInstance(self.v.geometry, dgpd.GeoSeries)
        self.assertTrue(
            (self.v.geometry == self.v.data.geometry).all().compute()
        )

    def test_tasks(self):
        assert hasattr(self.v, "tasks")
        v = self.v.copy()
        v._geo = v._geo.compute()
        assert not dask.is_dask_collection(v.data)
        assert v.tasks == 0
        v = self.v.to_lazy()
        assert v.tasks >= 1

    def test_bounds(self):
        self.assertTrue(hasattr(self.v, "bounds"))
        df = gpd.read_file("tests/data/vector/pods.shp")
        v = open_vectors("tests/data/vector/pods.shp")
        self.assertTrue(all(v.bounds == df.total_bounds))
        self.assertTrue(dask.is_dask_collection(v.to_lazy().bounds))
        self.assertTrue(all(v.to_lazy().bounds == df.total_bounds))


class TestSpecialMethods(unittest.TestCase):
    def setUp(self):
        self.v = open_vectors("tests/data/vector/pods.shp")

    def test_len(self):
        self.assertTrue(hasattr(self.v, "__len__"))
        self.assertTrue(len(self.v) == self.v.size)
        self.assertTrue(len(self.v) == len(self.v.data))

    def test_getitem(self):
        self.assertTrue(hasattr(self.v, "__getitem__"))
        self.assertIsInstance(self.v[0], Vector)
        self.assertTrue(
            self.v[0].data.compute().equals(self.v.data.loc[[0]].compute())
        )
        last = self.v.data.loc[[self.v.size - 1]]
        last.index = dask.array.from_array([0]).to_dask_dataframe()
        self.assertTrue(self.v[-1].data.compute().equals(last.compute()))
        with self.assertRaises(NotImplementedError):
            self.v[0:3]
        with self.assertRaises(TypeError):
            self.v[9.0]
        with self.assertRaises(IndexError):
            self.v[self.v.size]
        with self.assertRaises(IndexError):
            self.v[-200]
        with self.assertRaises(IndexError):
            self.v[999]


class TestCopy(unittest.TestCase):
    def test_copy(self):
        v = open_vectors("tests/data/vector/pods.shp")
        vc = v.copy()
        self.assertIsNot(v, vc)
        self.assertIsNot(v.data, vc.data)
        self.assertTrue(v.data.compute().equals(vc.data.compute()))


class TestLazyByDefault(unittest.TestCase):
    def test_lazy_by_default(self):
        v = open_vectors("tests/data/vector/pods.shp")
        self.assertTrue(dask.is_dask_collection(v._geo))


class TestEval(unittest.TestCase):
    def test_eval(self):
        v = open_vectors("tests/data/vector/Zones.gdb", 0)
        self.assertTrue(dask.is_dask_collection(v.data))
        ve = v.eval()
        self.assertTrue(dask.is_dask_collection(ve.data))
        self.assertTrue(ve.data.npartitions == 1)


class TestConversions(unittest.TestCase):
    def setUp(self):
        self.v = open_vectors("tests/data/vector/pods.shp")

    def test_to_lazy(self):
        vc = self.v.copy()
        vc._geo = vc._geo.compute()
        self.assertFalse(dask.is_dask_collection(vc.data))
        self.assertTrue(dask.is_dask_collection(vc.to_lazy().data))

    def test_to_dataframe(self):
        self.assertIsInstance(self.v.to_dataframe(), dgpd.GeoDataFrame)
        self.assertIsInstance(
            self.v.to_lazy().to_dataframe(), dgpd.GeoDataFrame
        )
        self.assertTrue(
            self.v.data.compute().equals(self.v.to_dataframe().compute())
        )

    def test_to_crs(self):
        crs = rio.crs.CRS.from_epsg(4326)
        self.assertTrue(self.v.crs == rio.crs.CRS.from_epsg(5070))
        self.assertTrue(self.v.to_crs(crs).crs == crs)
        self.assertTrue(self.v.to_crs(4326).crs == crs)
        self.assertTrue(self.v.to_crs("epsg:4326").crs == crs)


@pytest.mark.parametrize(
    "key", ["many_shapes", "single_shape", "one_part", "many_part"]
)
def test_to_raster(key):
    v = open_vectors("tests/data/vector/pods.shp")
    like = Raster("tests/data/raster/elevation.tif")
    if key == "many_shapes":
        truth = Raster("tests/data/raster/pods_like_elevation.tif")
        result = v.to_raster(like)
    elif key == "single_shape":
        truth = Raster("tests/data/raster/pods0_like_elevation.tif")
        result = v[0].to_raster(like)
    elif key == "one_part":
        truth = Raster("tests/data/raster/pods_like_elevation.tif")
        result = v.to_lazy().to_raster(like)
    elif key == "many_part":
        truth = Raster("tests/data/raster/pods_like_elevation.tif")
        v = Vector(v.to_lazy().data.repartition(10))
        result = v.to_raster(like)

    result_eval = result.eval()

    assert_valid_raster(result)
    assert result.null_value == 0
    assert result.dtype == np.dtype("uint8")
    assert result.shape[0] == 1
    assert result.crs == like.crs
    assert result.affine == like.affine
    assert result_eval.crs == like.crs
    assert result_eval.affine == like.affine
    assert np.allclose(result, truth)
    assert np.allclose(result.xdata.x, truth.xdata.x)
    assert np.allclose(result.xdata.y, truth.xdata.y)
    assert np.allclose(result.xdata.band, truth.xdata.band)
    if key == "single_shape":
        assert all(np.unique(result) == [0, 1])


def test_to_raster_chunks():
    # Make sure that chunking is preserved
    v = open_vectors("tests/data/vector/pods.shp")
    like = Raster("tests/data/raster/elevation.tif").chunk((1, 1000, 1000))
    truth = Raster("tests/data/raster/pods_like_elevation.tif")
    assert like.data.chunksize != like.shape

    result = v.to_raster(like)
    assert result.data.chunks == like.data.chunks
    assert np.allclose(result, truth)


def test_to_raster_field():
    v = open_vectors("tests/data/vector/pods.shp")
    like = Raster("tests/data/raster/elevation.tif")
    truth = Raster("tests/data/raster/pods_like_elevation_objectid_field.tif")
    result = v.to_raster(like, field="OBJECTID")

    result_eval = result.eval()

    assert_valid_raster(result)
    assert result.null_value == get_default_null_value(v.data.OBJECTID.dtype)
    assert result.dtype == v.data.OBJECTID.dtype
    assert result.shape[0] == 1
    assert result.crs == like.crs
    assert result.affine == like.affine
    assert result_eval.crs == like.crs
    assert result_eval.affine == like.affine
    assert result_eval.dtype == v.data.OBJECTID.dtype
    assert np.allclose(result, truth)
    assert np.allclose(result.xdata.x, truth.xdata.x)
    assert np.allclose(result.xdata.y, truth.xdata.y)
    assert np.allclose(result.xdata.band, truth.xdata.band)

    v = v.cast_field("OBJECTID", U16)
    result = v.to_raster(like, field="OBJECTID")
    assert result.dtype == I32
    assert result.null_value == get_default_null_value(I32)
    assert np.allclose(
        result, truth.set_null_value(get_default_null_value(I32))
    )


class TestCastField(unittest.TestCase):
    def test_cast_field(self):
        v = open_vectors("tests/data/vector/pods.shp")
        dtypes = [int, "int16", np.int32, float, "float"]
        for d in dtypes:
            dt = np.dtype(d)
            vc = v.cast_field("POD_Num", d)
            self.assertTrue(vc.data.POD_Num.dtype == dt)



from functools import reduce
from numbers import Integral, Number

import numpy as np
from xarray.core.dtypes import maybe_promote as xr_maybe_promote


def _add_type(mapping, dt, aliases):
    for d in [dt] + aliases:
        mapping[d] = dt


U8 = np.dtype(np.uint8)
U16 = np.dtype(np.uint16)
U32 = np.dtype(np.uint32)
U64 = np.dtype(np.uint64)
I8 = np.dtype(np.int8)
I16 = np.dtype(np.int16)
I32 = np.dtype(np.int32)
I64 = np.dtype(np.int64)
F16 = np.dtype(np.float16)
F32 = np.dtype(np.float32)
F64 = np.dtype(np.float64)
# Some machines don't support np.float128 so use longdouble instead. This
# aliases f128 on machines that support it and f64 on machines that don't
F128 = np.dtype(np.longdouble)
BOOL = np.dtype(bool)

DTYPE_INPUT_TO_DTYPE = {}
# Unsigned int
_add_type(DTYPE_INPUT_TO_DTYPE, U8, ["u1", "uint8", "B", np.uint8])
_add_type(DTYPE_INPUT_TO_DTYPE, U16, ["u2", "uint16", np.uint16])
_add_type(DTYPE_INPUT_TO_DTYPE, U32, ["u4", "uint32", np.uint32])
_add_type(DTYPE_INPUT_TO_DTYPE, U64, ["u8", "uint64", np.uint64])
# Signed int
_add_type(DTYPE_INPUT_TO_DTYPE, I8, ["i1", "int8", "b", "byte", np.int8])
_add_type(DTYPE_INPUT_TO_DTYPE, I16, ["i2", "int16", np.int16])
_add_type(DTYPE_INPUT_TO_DTYPE, I32, ["i4", "int32", "i", np.int32])
_add_type(DTYPE_INPUT_TO_DTYPE, I64, ["i8", "int64", "int", int, np.int64])
# Float
_add_type(DTYPE_INPUT_TO_DTYPE, F16, ["f2", "float16", np.float16])
_add_type(DTYPE_INPUT_TO_DTYPE, F32, ["f4", "float32", "f", np.float32])
_add_type(
    DTYPE_INPUT_TO_DTYPE,
    F64,
    ["f8", "float64", "d", "float", float, np.float64],
)
_add_type(DTYPE_INPUT_TO_DTYPE, F128, ["f16", "float128", np.longdouble])
# Boolean
_add_type(
    DTYPE_INPUT_TO_DTYPE,
    BOOL,
    ["bool", "?", bool, np.bool_, np.dtype("bool"), np.bool8],
)


def is_str(value):
    return isinstance(value, str)


def is_int(value_or_dtype):
    if isinstance(value_or_dtype, np.dtype):
        return value_or_dtype.kind in ("u", "i")
    return isinstance(value_or_dtype, Integral)


def is_float(value_or_dtype):
    if isinstance(value_or_dtype, np.dtype):
        return value_or_dtype.kind == "f"
    return is_scalar(value_or_dtype) and not is_int(value_or_dtype)


def is_bool(value_or_dtype):
    if isinstance(value_or_dtype, np.dtype):
        return value_or_dtype.kind == "b"
    return isinstance(value_or_dtype, (bool, np.bool_))


def is_scalar(value_or_dtype):
    if not isinstance(value_or_dtype, np.dtype):
        return isinstance(value_or_dtype, Number)
    return is_int(value_or_dtype) or is_float(value_or_dtype)


def get_common_dtype(values):
    return reduce(np.promote_types, map(np.min_scalar_type, values))


def promote_dtype_to_float(dtype):
    if dtype.kind == "f":
        return dtype
    if np.can_cast(dtype, F16):
        return F16
    elif np.can_cast(dtype, F32):
        return F32
    return F64


def maybe_promote(dtype):
    """Returns a dtype that can support missing values based on the input"""
    return xr_maybe_promote(dtype)[0]


def should_promote_to_fit(dtype, value):
    return is_float(value) and is_int(dtype)


def promote_data_dtype(xrs):
    dtype = maybe_promote(xrs.dtype)
    if dtype == xrs.dtype:
        return xrs
    return xrs.astype(dtype)



import numpy as np
import pytest

from raster_tools.distance import _heap as hp


def test_init_heap_data():
    keys, values, xrefs, heap_state = hp.init_heap_data(10, 100)
    assert keys.size == 32
    assert values.size == 16
    assert xrefs.size == 101
    assert (keys == np.inf).all()
    assert (values == -1).all()
    assert (xrefs == -1).all()
    assert heap_state[0]["count"] == 0

    keys, values, xrefs, heap_state = hp.init_heap_data(128, 1100)
    assert keys.size == 256
    assert values.size == 128
    assert xrefs.size == 1101
    assert (keys == np.inf).all()
    assert (values == -1).all()
    assert (xrefs == -1).all()
    assert heap_state[0]["count"] == 0

    keys, values, xrefs, heap_state = hp.init_heap_data(257, 1100)
    assert keys.size == 1024
    assert values.size == 512
    assert xrefs.size == 1101
    assert (keys == np.inf).all()
    assert (values == -1).all()
    assert (xrefs == -1).all()
    assert heap_state[0]["count"] == 0


@pytest.mark.parametrize(
    "capacity,max_value", [(-1, 100), (0, 100), (12, -1), (12, 0)]
)
def test_init_heap_data_errors(capacity, max_value):
    with pytest.raises(ValueError):
        hp.init_heap_data(capacity, max_value)
        hp.init_heap_data(capacity, max_value)


# TODO: Test the rest of the funcs



import sys
import warnings
from unittest import TestCase

import numpy as np
from scipy import stats

import raster_tools.stat_common as stc
from raster_tools.dtypes import is_scalar

PY_VER_37 = sys.version_info[0] == 3 and sys.version_info[1] == 7

INPUTS = [
    0,
    4.9,
    -1.0,
    np.array([]),
    np.array([0]),
    np.array([0, -1, 9]),
    np.arange(100, dtype=float),
    np.arange(600, dtype=float),
    np.arange(33_000, 0, -1, dtype=float),
    np.arange(66_000, 0, -1, dtype=float),
    np.array([np.nan, np.nan, np.nan]),
    np.array([np.nan, 0.0, np.nan]),
    np.array([np.nan, 1.0, 500.0]),
    np.array([np.nan, 1.0, 500.0]),
]


class TestJitStats(TestCase):
    def _test(self, func, arg, truth):
        res = func(arg)
        self.assertTrue(is_scalar(res))
        self.assertTrue(np.allclose(res, truth, equal_nan=True))

    def _test_func(self, func, truth_func):
        truth = []
        with warnings.catch_warnings():
            # Ignore nan related warnings
            warnings.simplefilter("ignore", category=RuntimeWarning)
            for i in INPUTS:
                try:
                    truth.append(truth_func(i))
                except ValueError:
                    truth.append(None)
        for i, t in zip(INPUTS, truth):
            if t is not None:
                self._test(func, i, t)
            else:
                with self.assertRaises(ValueError):
                    self._test(func, i, t)

    def test_nanmin_jit(self):
        self._test_func(stc.nanmin_jit, np.nanmin)

    def test_nanmax_jit(self):
        self._test_func(stc.nanmax_jit, np.nanmax)

    def test_nanmean_jit(self):
        self._test_func(stc.nanmean_jit, np.nanmean)

    def test_nanmedian_jit(self):
        self._test_func(stc.nanmedian_jit, np.nanmedian)

    def test_nansum_jit(self):
        self._test_func(stc.nansum_jit, np.nansum)

    def test_nanvar_jit(self):
        self._test_func(stc.nanvar_jit, np.nanvar)

    def test_nanstd_jit(self):
        self._test_func(stc.nanstd_jit, np.nanstd)

    def test_nan_unique_count_jit(self):
        def unique(x):
            uniq = np.unique(x)
            nans = np.isnan(uniq)
            uniq = uniq[~nans]
            return len(uniq)

        self._test_func(stc.nan_unique_count_jit, unique)

    def test_nanmode_jit(self):
        def mode(x):
            x = np.atleast_1d(x)
            x = x[~np.isnan(x)]
            if x.size == 0:
                return np.nan
            kwargs = {"nan_policy": "omit"}
            if not PY_VER_37:
                kwargs["keepdims"] = True
            m = stats.mode(x, **kwargs)
            return m[0][0]

        self._test_func(stc.nanmode_jit, mode)

    def test_nanentropy_jit(self):
        def entropy(x):
            x = np.atleast_1d(x)
            x = x[~np.isnan(x)]
            ent = 0.0
            _, cnts = np.unique(x, return_counts=True)
            n = len(cnts)
            if n:
                for c in cnts:
                    p = c / n
                    ent -= p * np.log(p)
            return ent

        self._test_func(stc.nanentropy_jit, entropy)

    def test_nanasm_jit(self):
        def asm(x):
            x = np.atleast_1d(x)
            x = x[~np.isnan(x)]
            ent = 0.0
            _, cnts = np.unique(x, return_counts=True)
            n = len(cnts)
            if n:
                for c in cnts:
                    p = c / n
                    ent += p * p
            return ent

        self._test_func(stc.nanasm_jit, asm)

    def test_nanargmin_jit(self):
        def argmin(x):
            x = np.atleast_1d(x)
            if x.size == 0:
                return -1
            if np.isnan(x).all():
                return -2
            return np.nanargmin(x)

        self._test_func(stc.nanargmin_jit, argmin)

    def test_nanargmax_jit(self):
        def argmax(x):
            x = np.atleast_1d(x)
            if x.size == 0:
                return -1
            if np.isnan(x).all():
                return -2
            return np.nanargmax(x)

        self._test_func(stc.nanargmax_jit, argmax)



import os
from pathlib import Path

import dask
import numpy as np
import rasterio as rio
import rioxarray as rxr
import xarray as xr
from dask.array.core import normalize_chunks as dask_chunks

from raster_tools.dtypes import F32, F64, I64, U8, is_bool, is_float, is_int
from raster_tools.masking import create_null_mask
from raster_tools.utils import (
    is_strictly_decreasing,
    is_strictly_increasing,
    validate_path,
)


class RasterIOError(BaseException):
    pass


class RasterDataError(BaseException):
    pass


def _get_extension(path):
    return os.path.splitext(path)[-1].lower()


def _get_chunking_info_from_file(src_file):
    with rio.open(src_file) as src:
        tile_shape = (1, *src.block_shapes[0])
        shape = (src.count, *src.shape)
        dtype = np.dtype(src.dtypes[0])
        return tile_shape, shape, dtype


def _get_chunks(data=None, src_file=None):
    chunks = (1, "auto", "auto")
    if data is None:
        if src_file is None:
            return chunks
        tile_shape, shape, dtype = _get_chunking_info_from_file(src_file)
        return dask_chunks(
            chunks, shape, dtype=dtype, previous_chunks=tile_shape
        )
    else:
        shape = data.shape
        dtype = data.dtype
        tile_shape = None
        if dask.is_dask_collection(data):
            tile_shape = data.chunks
        elif src_file is not None:
            _, tile_shape, _ = _get_chunking_info_from_file(src_file)
        return dask_chunks(
            chunks, shape, dtype=dtype, previous_chunks=tile_shape
        )


def chunk(xrs, src_file=None):
    if isinstance(xrs, xr.Dataset):
        chunks = _get_chunks(xrs.raster, src_file)
        return xrs.chunk({d: c for d, c in zip(xrs.raster.dims, chunks)})
    else:
        return xrs.chunk(_get_chunks(xrs, src_file))


TIFF_EXTS = frozenset((".tif", ".tiff"))
NC_EXTS = frozenset((".cdf", ".nc", ".nc4"))
HDF_EXTS = frozenset((".hdf", ".h4", ".hdf4", ".he2", ".h5", ".hdf5", ".he5"))
GRIB_EXTS = frozenset((".grib", ".grib2", ".grb", ".grb2", ".gb", ".gb2"))
BATCH_EXTS = frozenset((".bch",))

# File extenstions that can't be read in yet
READ_NOT_IMPLEMENTED_EXTS = NC_EXTS | HDF_EXTS | GRIB_EXTS
# File extenstions that can't be written out yet
WRITE_NOT_IMPLEMENTED_EXTS = NC_EXTS | HDF_EXTS | GRIB_EXTS

IO_UNDERSTOOD_TYPES = (str, Path)


def is_batch_file(path):
    return _get_extension(path) in BATCH_EXTS


def normalize_xarray_data(xrs):
    if len(xrs.shape) > 3 or len(xrs.shape) < 2:
        raise ValueError(
            "Invalid shape. xarray.DataArray objects must have 2D or 3D "
            "shapes."
        )
    if len(xrs.shape) == 2:
        # Add band dim
        xrs = xrs.expand_dims({"band": [1]})
    dims = xrs.dims
    if "lon" in dims:
        xrs = xrs.rename({"lon": "x"})
        dims = xrs.dims
    if "lat" in dims:
        xrs = xrs.rename({"lat": "y"})
        dims = xrs.dims
    if not dims == ("band", "y", "x"):
        # No easy way to figure out how best to transpose based on dim names so
        # just assume the order is valid and rename.
        xrs = xrs.rename(
            {
                d: new_d
                for d, new_d in zip(dims, ("band", "y", "x"))
                if d != new_d
            }
        )
    if xrs.band.values[0] != 1:
        xrs["band"] = np.arange(1, len(xrs.band) + 1)
    if any(dim not in xrs.coords for dim in xrs.dims):
        raise ValueError(
            "Invalid coordinates on xarray.DataArray object:\n{xrs!r}"
        )
    # Make sure that x and y are always increasing. xarray will auto align
    # rasters but when a raster is converted to a numpy or dask array, the
    # data may not be aligned. This ensures that rasters converted to
    # non-georeferenecd formats will be oriented the same.
    if is_strictly_decreasing(xrs.x):
        xrs = xrs.isel(x=slice(None, None, -1))
    if is_strictly_increasing(xrs.y):
        xrs = xrs.isel(y=slice(None, None, -1))
    tf = xrs.rio.transform(True)
    xrs = xrs.rio.write_transform(tf)
    return xrs


ESRI_DEFAULT_F32_NV = np.finfo(F32).min


def normalize_null_value(nv, dtype):
    # Make sure that ESRI's default F32 null value is properly
    # registered as F32
    if dtype == F32 and nv is not None and np.isclose(nv, ESRI_DEFAULT_F32_NV):
        nv = F32.type(nv)
    # Some rasters have (u)int dtype and a null value that is a whole number
    # but it gets read in as a float. This can cause a lot of accidental type
    # promotions down the pipeline. Check for this case and correct it.
    if is_int(dtype) and is_float(nv) and float(nv).is_integer():
        nv = int(nv)
    return nv


def open_raster_from_path(path):
    if type(path) in IO_UNDERSTOOD_TYPES:
        path = str(path)
        path = os.path.abspath(path)
    else:
        raise RasterIOError(
            f"Could not resolve input to a raster path: '{path}'"
        )
    validate_path(path)
    ext = _get_extension(path)

    xrs = None
    # Try to let gdal open anything but NC, HDF, GRIB files
    if not ext or ext not in READ_NOT_IMPLEMENTED_EXTS:
        try:
            xrs = rxr.open_rasterio(path, chunks=_get_chunks())
        except rio.errors.RasterioIOError as e:
            raise RasterIOError(str(e))
    elif ext in READ_NOT_IMPLEMENTED_EXTS:
        raise NotImplementedError(
            "Reading of NetCDF, HDF, and GRIB files is not supported at this"
            " time."
        )
    else:
        raise RasterIOError("Unknown file type")
    if isinstance(xrs, xr.Dataset):
        raise RasterDataError("Too many data variables in input data")
    assert isinstance(
        xrs, xr.DataArray
    ), "Resulting data structure must be a DataArray"
    if not dask.is_dask_collection(xrs):
        xrs = chunk(xrs, path)

    xrs = normalize_xarray_data(xrs)

    nv = xrs.attrs.get("_FillValue", None)
    nv = normalize_null_value(nv, xrs.dtype)
    mask = create_null_mask(xrs, nv)
    return xrs, mask, nv


def write_raster(xrs, path, no_data_value, **rio_gdal_kwargs):
    ext = _get_extension(path)
    rio_is_bool = False
    if ext in TIFF_EXTS or len(ext) == 0:
        if xrs.dtype == I64:
            # GDAL, and thus rioxarray and rasterio, doesn't support I64 so
            # cast up to float. This avoids to_raster throwing a TypeError.
            xrs = xrs.astype(F64)
        elif is_bool(xrs.dtype):
            # GDAL doesn't support boolean dtype either so convert to uint8
            # 0-1 encoding.
            rio_is_bool = True
            xrs = xrs.astype(U8)

    if not ext or ext not in WRITE_NOT_IMPLEMENTED_EXTS:
        kwargs = {"lock": True, "compute": True, **rio_gdal_kwargs}
        if "blockheight" in kwargs:
            value = kwargs.pop("blockheight")
            kwargs["blockysize"] = value
        if "blockwidth" in kwargs:
            value = kwargs.pop("blockwidth")
            kwargs["blockxsize"] = value
        if rio_is_bool:
            # Store each entry using a single bit
            kwargs["nbits"] = 1
        xrs.rio.to_raster(path, **kwargs)
    else:
        # TODO: populate
        raise NotImplementedError()



import warnings

import dask
import numpy as np

from raster_tools.dtypes import (
    BOOL,
    DTYPE_INPUT_TO_DTYPE,
    F16,
    F32,
    F64,
    F128,
    I8,
    I16,
    I32,
    I64,
    U8,
    U16,
    U32,
    U64,
    is_bool,
    is_float,
)


def create_null_mask(xrs, null_value):
    if null_value is not None:
        if not np.isnan(null_value):
            mask = xrs.data == null_value
        else:
            mask = np.isnan(xrs.data)
    else:
        mask = dask.array.zeros_like(xrs, dtype=bool)
    return mask


DTYPE_TO_DEFAULT_NULL = {
    BOOL: True,
    U8: U8.type(99),
    U16: U16.type(9_999),
    U32: U32.type(999_999_999),
    U64: U64.type(999_999_999_999),
    I8: I8.type(-99),
    I16: I16.type(-9_999),
    I32: I32.type(-999_999),
    I64: I64.type(-999_999),
    F16: np.finfo(F16).min,
    F32: F32.type(-999_999.0),
    F64: F64.type(-999_999.0),
    F128: F128.type(-999_999.0),
}


def get_default_null_value(dtype):
    """Get the default null value for a given dtype."""
    dtype = DTYPE_INPUT_TO_DTYPE[dtype]
    return DTYPE_TO_DEFAULT_NULL[dtype]


def reconcile_nullvalue_with_dtype(null_value, dtype, warn=False):
    """Make sure that the null value is consistent with the given dtype

    The null value is cast to the given dtype, if possible. If not, a warning
    is raised and the default for the dtype is returned.
    """
    if null_value is None:
        return None
    if np.isnan(null_value):
        return get_default_null_value(dtype)

    dtype = np.dtype(dtype)
    if np.can_cast(null_value, dtype):
        return null_value

    original_nv = null_value
    if is_float(original_nv) and float(original_nv).is_integer():
        nv = int(original_nv)
    else:
        nv = original_nv
    if np.can_cast(nv, dtype) or (is_bool(dtype) and nv in (0, 1)):
        return dtype.type(nv)
    nv = get_default_null_value(dtype)
    if warn:
        warnings.warn(
            f"The null value {original_nv!r} could not be cast to {dtype}. "
            f"It has been automatically changed to {nv!r}",
            stacklevel=2,
        )
    return nv



from .cost_distance import (
    cda_allocation,
    cda_cost_distance,
    cda_traceback,
    cost_distance_analysis,
)
from .proximity import (
    pa_allocation,
    pa_direction,
    pa_proximity,
    proximity_analysis,
)

__all__ = [
    "cda_allocation",
    "cda_cost_distance",
    "cda_traceback",
    "cost_distance_analysis",
    "pa_allocation",
    "pa_direction",
    "pa_proximity",
    "proximity_analysis",
]



from functools import partial

import dask.array as da
import numba as nb
import numpy as np
from dask_image import ndfilters

from raster_tools.dtypes import (
    F64,
    U8,
    U16,
    U32,
    U64,
    is_bool,
    is_float,
    is_int,
    promote_data_dtype,
    promote_dtype_to_float,
)
from raster_tools.raster import get_raster
from raster_tools.stat_common import (
    nan_unique_count_jit,
    nanasm_jit,
    nanentropy_jit,
    nanmax_jit,
    nanmean_jit,
    nanmedian_jit,
    nanmin_jit,
    nanmode_jit,
    nanstd_jit,
    nansum_jit,
    nanvar_jit,
)

__all__ = [
    "check_kernel",
    "convolve",
    "correlate",
    "focal",
    "get_focal_window",
]


ngjit = nb.jit(nopython=True, nogil=True)


@ngjit
def _focal_chunk(chunk, kernel, func):
    out = np.empty_like(chunk)
    rows, cols = chunk.shape
    krows, kcols = kernel.shape
    kr_top = (krows - 1) // 2
    kr_bot = krows // 2
    kc_left = (kcols - 1) // 2
    kc_right = kcols // 2
    r_ovlp = max(kr_top, kr_bot)
    c_ovlp = max(kc_left, kc_right)
    kernel_values = np.empty(kernel.size, dtype=chunk.dtype)
    kernel = kernel.ravel()

    # iterate over rows, skipping outer overlap rows
    for r in range(r_ovlp, rows - r_ovlp):
        # iterate over cols, skipping outer overlap cols
        for c in range(c_ovlp, cols - c_ovlp):
            kernel_values.fill(np.nan)
            # Kernel flat index
            ki = 0
            # iterate over kernel footprint, this extends into overlap regions
            # at edges
            for kr in range(r - kr_top, r + kr_bot + 1):
                for kc in range(c - kc_left, c + kc_right + 1):
                    if kernel[ki]:
                        kernel_values[ki] = chunk[kr, kc]
                    ki += 1
            out[r, c] = func(kernel_values)
    return out


@nb.jit(nopython=True, nogil=True, parallel=True)
def _correlate2d_chunk(chunk, kernel):
    out = np.empty_like(chunk)
    rows, cols = chunk.shape
    krows, kcols = kernel.shape
    kr_top = (krows - 1) // 2
    kr_bot = krows // 2
    kc_left = (kcols - 1) // 2
    kc_right = kcols // 2
    r_ovlp = max(kr_top, kr_bot)
    c_ovlp = max(kc_left, kc_right)
    kernel = kernel.ravel()

    # iterate over rows, skipping outer overlap rows
    for r in nb.prange(r_ovlp, rows - r_ovlp):
        # iterate over cols, skipping outer overlap cols
        for c in nb.prange(c_ovlp, cols - c_ovlp):
            # Kernel flat index
            ki = 0
            # iterate over kernel footprint, this extends into overlap regions
            # at edges
            v = 0.0
            for kr in range(r - kr_top, r + kr_bot + 1):
                for kc in range(c - kc_left, c + kc_right + 1):
                    val = chunk[kr, kc]
                    if not np.isnan(val):
                        v += kernel[ki] * chunk[kr, kc]
                    ki += 1
            out[r, c] = v
    return out


@ngjit
def _get_offsets(kernel_shape):
    """
    Returns the number of cells on either side of kernel center in both
    directions.
    """
    krows, kcols = kernel_shape
    kr_top = (krows - 1) // 2
    kr_bot = krows // 2
    kc_left = (kcols - 1) // 2
    kc_right = kcols // 2
    return ((kr_top, kr_bot), (kc_left, kc_right))


def _focal_dask_map(data, chunk_func, kernel, boundary=np.nan):
    offsets = _get_offsets(kernel.shape)
    # map_overlap does not support asymmetrical padding so take max. This adds
    # at most one extra pixel to each dim.
    rpad = max(offsets[0])
    cpad = max(offsets[1])
    return data.map_overlap(
        chunk_func,
        depth={0: rpad, 1: cpad},
        boundary=boundary,
        dtype=data.dtype,
        meta=np.array((), dtype=data.dtype),
    )


def _focal_dispatch(
    operation, data, kernel, kernel_func=None, boundary=np.nan
):
    # TODO: check for cupy eventually
    if operation == "correlate":
        chunk_func = partial(_correlate2d_chunk, kernel=kernel)
    else:
        chunk_func = partial(_focal_chunk, kernel=kernel, func=kernel_func)
    return _focal_dask_map(data, chunk_func, kernel, boundary=boundary)


def check_kernel(kernel):
    """Validate and return the focal kernel.

    Parameters
    ----------
    kernel : 2D numpy.ndarray
        The kernel to validate. Errors are raised if it is not a numpy array,
        contains nan values, or is not 2D.

    Returns
    -------
    numpy.ndarray
        The input `kernel`, if nothing was wrong with it.

    """
    if not isinstance(kernel, np.ndarray):
        raise TypeError("Kernel must be numpy.ndarray")
    if len(kernel.shape) != 2:
        raise ValueError("Kernel must be 2D")
    if np.isnan(kernel).any():
        raise ValueError("Kernel can't contain NaN values")
    return kernel


def _check_data(data):
    if not isinstance(data, (np.ndarray, da.Array)):
        raise TypeError("Kernel must be numpy.ndarray or dask.array.Array")
    if len(data.shape) != 2:
        raise ValueError("Data must be 2D")


_MODE_TO_DASK_BOUNDARY = {
    "reflect": "reflect",
    "nearest": "nearest",
    "wrap": "periodic",
    "constant": "constant",
}
_VALID_CORRELATE_MODES = frozenset(_MODE_TO_DASK_BOUNDARY.keys())


def _correlate(data, kernel, mode="constant", cval=0.0, nan_aware=False):
    """Cross-correlates a `kernel` with `data`.

    This function can be used for convolution as well; just rotate the kernel
    180 degress (e.g. ``kernel = kernel[::-1, ::-1])`` before calling this
    function.

    Parameters
    ----------
    data : 2D ndarray or dask array
        Data to cross-correlate the kernel with.
    kernel : 2D ndarray
        Kernel to apply to the data through cross-correlation.
    mode : {'reflect', 'nearest', 'wrap', 'constant'}, optional
        The mode to use for the edges of the data. Default is 'constant'.
    cval : scalar, optional
        The value to use when `mode` is 'constant'. Default is ``0.0``.
    nan_aware : bool, optional
        If ``True``, NaN values are ignored during correlation. If ``False``,
        a faster correlation algorithm can be used. Default is ``False``.

    Returns
    -------
    correlated : 2D dask array
        The cross-correlation result as a lazy dask array.

    """
    _check_data(data)
    check_kernel(kernel)
    if mode not in _VALID_CORRELATE_MODES:
        raise ValueError(f"Invalid mode: '{mode}'")
    if isinstance(data, np.ndarray):
        data = da.from_array(data)
    if nan_aware:
        data = promote_data_dtype(data)
    if is_bool(kernel.dtype):
        kernel = kernel.astype(int)
    if is_float(data.dtype) and is_int(kernel.dtype):
        kernel = kernel.astype(data.dtype)
    if is_int(data.dtype) and is_float(kernel.dtype):
        data = promote_data_dtype(data)

    if nan_aware:
        boundary = _MODE_TO_DASK_BOUNDARY[mode]
        if boundary == "constant":
            boundary = cval
        return _focal_dispatch("correlate", data, kernel, boundary=boundary)
    # Shift pixel origins to match ESRI behavior for even shaped kernels
    shift_origin = [d % 2 == 0 for d in kernel.shape]
    origin = [-1 if shift else 0 for shift in shift_origin]
    return ndfilters.correlate(
        data, kernel, mode=mode, cval=cval, origin=origin
    )


FOCAL_STATS = frozenset(
    (
        "asm",
        "entropy",
        "max",
        "mean",
        "median",
        "mode",
        "min",
        "std",
        "sum",
        "var",
        "unique",
    )
)
_STAT_TO_FUNC = {
    "asm": nanasm_jit,
    "entropy": nanentropy_jit,
    "max": nanmax_jit,
    "mean": nanmean_jit,
    "median": nanmedian_jit,
    "mode": nanmode_jit,
    "min": nanmin_jit,
    "std": nanstd_jit,
    "sum": nansum_jit,
    "var": nanvar_jit,
    "unique": nan_unique_count_jit,
}


def _focal(data, kernel, stat, nan_aware=False):
    """Apply a focal stat function.

    Applies the `stat` function to the `data` using `kernel` to determine the
    neighborhood for each pixel. `nan_aware` indicates whether the filter
    should handle NaN values. If `nan_aware` is False, optimizations may be
    made.

    """
    if stat not in FOCAL_STATS:
        raise ValueError(f"Unknown focal stat: '{stat}'")
    _check_data(data)
    check_kernel(kernel)

    if isinstance(data, np.ndarray):
        data = da.from_array(data)

    kernel = kernel.astype(bool)
    if not nan_aware and stat in ("min", "max", "sum"):
        # Use ndfilters which is faster but dosn't handle nan values
        if stat == "min":
            func = partial(
                ndfilters.minimum_filter, footprint=kernel, mode="nearest"
            )
        elif stat == "max":
            func = partial(
                ndfilters.maximum_filter, footprint=kernel, mode="nearest"
            )
        elif stat == "sum":
            func = partial(
                ndfilters.correlate, weights=kernel, mode="constant"
            )
        return func(data)
    # Use _focal_dispatch which is slower but handles nan values.
    # Promote to allow for nan boundary fill values.
    new_dtype = promote_dtype_to_float(data.dtype)
    if new_dtype != data.dtype:
        data = data.astype(new_dtype)
    return _focal_dispatch("focal", data, kernel, _STAT_TO_FUNC[stat])


def focal(raster, focal_type, width_or_radius, height=None):
    """Applies a focal filter to raster bands individually.

    The filter uses a window/footprint that is created using the
    `width_or_radius` and `height` parameters. The window can be a
    rectangle, circle or annulus.

    Parameters
    ----------
    raster : Raster or path str
        The raster to perform the focal operation on.
    focal_type : str
        Specifies the aggregation function to apply to the focal
        neighborhood at each pixel. Can be one of the following string
        values:

        'min'
            Finds the minimum value in the neighborhood.
        'max'
            Finds the maximum value in the neighborhood.
        'mean'
            Finds the mean of the neighborhood.
        'median'
            Finds the median of the neighborhood.
        'mode'
            Finds the mode of the neighborhood.
        'sum'
            Finds the sum of the neighborhood.
        'std'
            Finds the standard deviation of the neighborhood.
        'var'
            Finds the variance of the neighborhood.
        'asm'
            Angular second moment. Applies -sum(P(g)**2) where P(g) gives
            the probability of g within the neighborhood.
        'entropy'
            Calculates the entropy. Applies -sum(P(g) * log(P(g))). See
            'asm' above.
        'unique'
            Calculates the number of unique values in the neighborhood.
    width_or_radius : int or 2-tuple of ints
        If an int and `height` is `None`, specifies the radius of a circle
        window. If an int and `height` is also an int, specifies the width
        of a rectangle window. If a 2-tuple of ints, the values specify the
        inner and outer radii of an annulus window.
    height : int or None
        If `None` (default), `width_or_radius` will be used to construct a
        circle or annulus window. If an int, specifies the height of a
        rectangle window.

    Returns
    -------
    Raster
        The resulting raster with focal filter applied to each band. The
        bands will have the same shape as the original Raster.

    """
    raster = get_raster(raster)
    if focal_type not in FOCAL_STATS:
        raise ValueError(f"Unknown focal operation: '{focal_type}'")

    window = get_focal_window(width_or_radius, height)
    rs = raster.copy()
    data = rs.data

    # Convert to float and fill nulls with nan, if needed
    if raster._masked:
        new_dtype = promote_dtype_to_float(raster.dtype)
        if new_dtype != data.dtype:
            data = data.astype(new_dtype)
        data = da.where(raster.mask, np.nan, data)

    result = [
        _focal(data[bnd], window, focal_type, raster._masked)
        for bnd in range(data.shape[0])
    ]
    data = da.stack(result)

    if raster._masked:
        if focal_type == "unique":
            # Values of 0 mean that all values in window were nan (AKA null) so
            # replace them with nan
            data = da.where(data == 0, np.nan, data)
    else:
        if focal_type == "mode" and is_int(raster.dtype):
            data = data.astype(raster.dtype)
        elif focal_type == "unique":
            n = window.size
            unq_dtype = None
            for dt in (U8, U16, U32, U64):
                if np.can_cast(n, dt):
                    unq_dtype = dt
                    break
            data = data.astype(unq_dtype)

    # Nan values will only appear in the result data if there were null values
    # present in the input. Thus we only need to worry about updating the mask
    # if the input was masked.
    if raster._masked:
        rs.xmask.data = raster.mask | np.isnan(data)
        data = da.where(rs.mask, rs.null_value, data)
    rs.xdata.data = data
    return rs


def get_focal_window(width_or_radius, height=None):
    """Get a rectangle, circle, or annulus focal window.

    A rectangle window is simply a NxM grid of ``True`` values. A circle window
    is a grid with a centered circle of ``True`` values surrounded by `False`
    values. The circle extends to the edge of the grid. An annulus window is
    the same as a circle window but  has a nested circle of `False` values
    inside the main circle.

    Parameters
    ----------
    width_or_radius : int or 2-tuple of ints
        If an int and `height` is `None`, specifies the radius of a circle
        window. If an int and `height` is also an int, specifies the width of
        a rectangle window. If a 2-tuple of ints, the values specify the inner
        and outer radii of an annulus window.
    height : int or None
        If `None` (default), `width_or_radius` will be used to construct a
        circle or annulus window. If an int, specifies the height of a
        rectangle window.

    Returns
    -------
    window : ndarray
        A focal window containing bools with the specified pattern.

    """
    if isinstance(width_or_radius, (list, tuple)):
        if len(width_or_radius) != 2:
            raise ValueError(
                "If width_or_radius is a sequence, it must be size 2"
            )
        if width_or_radius[0] >= width_or_radius[1]:
            raise ValueError(
                "First radius value must be less than or equal to the second."
            )
    else:
        width_or_radius = [width_or_radius]
    for value in width_or_radius:
        if not is_int(value):
            raise TypeError(
                f"width_or_radius values must be integers: {value}"
            )
        elif value <= 0:
            raise ValueError(
                "Window width or radius values must be greater than 0."
                f" Got {value}"
            )
    if height is not None:
        if len(width_or_radius) == 2:
            raise ValueError(
                "height must be None if width_or_radius indicates annulus"
            )
        if not is_int(height):
            raise TypeError(f"height must be an integer or None: {height}")
        elif height <= 0:
            raise ValueError(
                f"Window height must be greater than 0. Got {height}"
            )

    window_out = None
    windows = []
    if height is None:
        for rvalue in width_or_radius:
            width = ((rvalue - 1) * 2) + 1
            height = width
            r = (width - 1) // 2
            window = np.zeros((height, width), dtype=bool)
            for x in range(width):
                for y in range(height):
                    rxy = np.sqrt((x - r) ** 2 + (y - r) ** 2)
                    if rxy <= r:
                        window[x, y] = True
            windows.append(window)
        if len(windows) != 2:
            window_out = windows[0]
        else:
            w1, w2 = windows
            padding = (np.array(w2.shape) - np.array(w1.shape)) // 2
            w1 = np.pad(w1, padding, mode="constant", constant_values=False)
            w2[w1] = False
            window_out = w2
    else:
        width = width_or_radius[0]
        window_out = np.ones((width, height), dtype=bool)
    return window_out


def correlate(raster, kernel, mode="constant", cval=0.0):
    """Cross-correlate `kernel` with each band individually. Returns a new
    Raster.

    The kernel is applied to each band in isolation so returned raster has
    the same shape as the original.

    Parameters
    ----------
    raster : Raster or path str
        The raster to cross-correlate `kernel` with. Can be multibanded.
    kernel : array_like
        2D array of kernel weights
    mode : {'reflect', 'constant', 'nearest', 'wrap'}, optional
        Determines how the data is extended beyond its boundaries. The
        default is 'constant'.

        'reflect' (d c b a | a b c d | d c b a)
            The data pixels are reflected at the boundaries.
        'constant' (k k k k | a b c d | k k k k)
            A constant value determined by `cval` is used to extend the
            data pixels.
        'nearest' (a a a a | a b c d | d d d d)
            The data is extended using the boundary pixels.
        'wrap' (a b c d | a b c d | a b c d)
            The data is extended by wrapping to the opposite side of the
            grid.
    cval : scalar, optional
        Value used to fill when `mode` is 'constant'. Default is 0.0.

    Returns
    -------
    Raster
        The resulting new Raster.
    """
    raster = get_raster(raster)
    kernel = np.asarray(kernel)
    check_kernel(kernel)
    rs = raster.copy()
    if is_float(kernel.dtype) and is_int(rs.dtype):
        rs.xdata.data = rs.data.astype(F64)
    data = rs.data
    final_dtype = data.dtype

    # Convert to float and fill nulls with nan, if needed
    upcast = False
    if raster._masked:
        new_dtype = promote_dtype_to_float(data.dtype)
        upcast = new_dtype != data.dtype
        if upcast:
            data = data.astype(new_dtype)
        data = da.where(raster.mask, np.nan, data)

    for bnd in range(data.shape[0]):
        data[bnd] = _correlate(
            data[bnd], kernel, mode=mode, cval=cval, nan_aware=rs._masked
        )

    # Cast back to int, if needed
    if upcast:
        data = data.astype(final_dtype)

    rs.xdata.data = data
    return rs.burn_mask()


def convolve(raster, kernel, mode="constant", cval=0.0):
    """Convolve `kernel` with each band individually. Returns a new Raster.

    This is the same as correlation but the kernel is rotated 180 degrees,
    e.g. ``kernel = kernel[::-1, ::-1]``.  The kernel is applied to each
    band in isolation so the returned raster has the same shape as the
    original.

    Parameters
    ----------
    raster : Raster
        The raster to convolve `kernel` with. Can be multibanded.
    kernel : array_like
        2D array of kernel weights
    mode : {'reflect', 'constant', 'nearest', 'wrap'}, optional
        Determines how the data is extended beyond its boundaries. The
        default is 'constant'.

        'reflect' (d c b a | a b c d | d c b a)
            The data pixels are reflected at the boundaries.
        'constant' (k k k k | a b c d | k k k k)
            A constant value determined by `cval` is used to extend the
            data pixels.
        'nearest' (a a a a | a b c d | d d d d)
            The data is extended using the boundary pixels.
        'wrap' (a b c d | a b c d | a b c d)
            The data is extended by wrapping to the opposite side of the
            grid.
    cval : scalar, optional
        Value used to fill when `mode` is 'constant'. Default is 0.0.

    Returns
    -------
    Raster
        The resulting new Raster.
    """
    kernel = np.asarray(kernel)
    check_kernel(kernel)
    kernel = kernel[::-1, ::-1].copy()
    return correlate(raster, kernel, mode=mode, cval=cval)



# isort: off
# TODO(pygeos): remove this once shapely is the default backend for geopandas.
# Force raster_tools._compat to be loaded before geopandas when running tests
import raster_tools  # noqa: F401

# isort: on

from unittest import TestCase

import dask
import dask_geopandas as dgpd
import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import shapely

from raster_tools import Raster, band_concat, open_vectors
from raster_tools.zonal import (
    _ZONAL_STAT_FUNCS,
    ZONAL_STAT_FUNCS,
    extract_points_eager,
    zonal_stats,
)
from tests.test_focal import asm, entropy, mode, unique
from tests.utils import arange_raster


def _apply_stat(dem, vrs, func):
    rsnp = np.array(vrs)
    mask = rsnp > 0
    return func(dem[mask])


def _apply_stat_all(dem, vrss, func):
    return [_apply_stat(dem, vrs, func) for vrs in vrss]


def count(x):
    return np.count_nonzero(~np.isnan(x))


all_stats = {
    "asm": asm,
    "count": count,
    "entropy": entropy,
    "max": np.nanmax,
    "mean": np.nanmean,
    "median": np.nanmedian,
    "min": np.nanmin,
    "mode": mode,
    "std": np.nanstd,
    "sum": np.nansum,
    "unique": unique,
    "var": np.nanvar,
}


class TestZonalStats(TestCase):
    def setUp(self):
        self.dem = Raster("tests/data/raster/elevation.tif")
        self.dem_np = np.array(self.dem)
        self.vc = open_vectors("tests/data/vector/pods_first_10.shp")
        self.vc_rasters = [v.to_raster(self.dem).eval() for v in self.vc]

    def test_stat_func_set(self):
        # Make sure public set of funcs matches private look up table
        self.assertTrue(ZONAL_STAT_FUNCS == set(_ZONAL_STAT_FUNCS))

    def test_zonal_stats(self):
        res = zonal_stats(self.vc, self.dem, "mode")
        resc = res.compute()
        # Check that result is lazy
        self.assertTrue(dask.is_dask_collection(res))

        mode_truth = _apply_stat_all(self.dem_np, self.vc_rasters, mode)
        self.assertTrue(all(resc.columns == ["band", "mode"]))
        self.assertTrue(len(resc["mode"]) == 10)
        self.assertTrue(all(resc.index == list(range(10))))
        self.assertTrue(np.allclose(resc["mode"], mode_truth))

        stats = {"std": np.nanstd, "count": count, "asm": asm}
        resc = zonal_stats(self.vc, self.dem, list(stats)).compute()

        truth = {
            k: _apply_stat_all(self.dem_np, self.vc_rasters, func)
            for k, func in stats.items()
        }
        self.assertTrue(all(resc.columns == ["band", *stats]))
        for k in stats:
            self.assertTrue(np.allclose(resc[k], truth[k], 1e5))

    def test_lazy(self):
        res = zonal_stats(self.vc.to_lazy(), self.dem, "mode")
        # Check that result is lazy
        self.assertTrue(dask.is_dask_collection(res))
        mode_truth = _apply_stat_all(self.dem_np, self.vc_rasters, mode)
        self.assertTrue(np.allclose(res["mode"], mode_truth))

    def test_all_stats(self):
        self.assertTrue(set(all_stats) == set(ZONAL_STAT_FUNCS))
        res = zonal_stats(self.vc, self.dem, list(all_stats)).compute()

        truth = {
            k: _apply_stat_all(self.dem_np, self.vc_rasters, func)
            for k, func in all_stats.items()
        }
        self.assertTrue(all(res.columns == ["band", *all_stats]))
        for k in all_stats:
            self.assertTrue(np.allclose(res[k], truth[k], 1e5))

    def test_multiband_all_stats(self):
        dem = band_concat((self.dem, self.dem * 105))
        dem_np = np.array(dem)
        res = zonal_stats(self.vc, dem, list(all_stats)).compute()

        truth = {
            k: _apply_stat_all(dem_np[0:1], self.vc_rasters, func)
            for k, func in all_stats.items()
        }
        for k in truth:
            truth[k] = np.concatenate(
                [
                    truth[k],
                    _apply_stat_all(dem_np[1:], self.vc_rasters, all_stats[k]),
                ]
            )
        self.assertTrue(all(res.columns == ["band", *all_stats]))
        band = np.ones(2 * len(self.vc))
        band[len(self.vc) :] = 2
        self.assertTrue(len(res) == 2 * len(self.vc))
        self.assertTrue(np.allclose(res.band, band))
        for k in all_stats:
            self.assertTrue(np.allclose(res[k], truth[k], 1e5))


def get_random_points(n, nparts, dem):
    xmin, ymin, xmax, ymax = dem.bounds
    xspan = xmax - xmin
    yspan = ymax - ymin
    x = xmin + (xspan * 1.1 * np.random.random(n)) - (0.05 * xspan)
    y = ymin + (yspan * 1.1 * np.random.random(n)) - (0.05 * yspan)
    df = gpd.GeoSeries.from_xy(x, y, crs=dem.crs).to_frame("geometry")
    return dgpd.from_geopandas(df, npartitions=nparts)


def dem_clipped_small():
    return Raster("tests/data/raster/elevation_clipped_small.tif")


@pytest.mark.parametrize(
    "dem,n,nparts,name",
    [
        (dem_clipped_small(), 1_000, 1, None),
        (dem_clipped_small(), 1_000, 4, "values"),
        (dem_clipped_small().chunk((1, 10, 10)), 1_000, 4, "points"),
        (
            band_concat(
                [
                    dem_clipped_small(),
                    dem_clipped_small() + 1,
                    dem_clipped_small() + 10,
                ]
            ).chunk((1, 10, 10)),
            1_000,
            4,
            None,
        ),
        (arange_raster((700, 6, 6)).set_crs("EPSG:3857"), 10, 1, None),
    ],
)
def test_extract_points_eager(dem, n, nparts, name):
    if name is None:
        name = "extracted"
    df = get_random_points(n, nparts, dem)
    points = df.geometry.compute().to_list()
    x, y = dask.compute(df.geometry.x, df.geometry.y)
    r, c = dem.index(x, y)
    data = dem.values
    mask = dem.mask.compute()
    nb, nr, nc = data.shape
    valid = (r >= 0) & (r < nr) & (c >= 0) & (c < nc)
    n = len(x)
    bbox = shapely.geometry.box(*dem.bounds)
    point_check = []
    for p in points:
        point_check.append(bbox.contains(p))
    assert np.allclose(valid, point_check)
    dfs = []
    for bnd in range(nb):
        d = {
            "band": np.full(n, bnd + 1, dtype=np.min_scalar_type(bnd + 1)),
            name: np.full(n, np.nan),
        }
        extracted = data[bnd, r[valid], c[valid]].astype(float)
        masked = mask[bnd, r[valid], c[valid]]
        extracted[masked] = np.nan
        d[name][valid] = extracted
        dfs.append(
            pd.DataFrame(d, index=pd.RangeIndex(n * bnd, n * (bnd + 1)))
        )
    truth = pd.concat(dfs)
    result = extract_points_eager(df, dem, name)
    resultc = result.compute()
    assert isinstance(result, dask.dataframe.DataFrame)
    assert isinstance(resultc, pd.DataFrame)
    assert result.band.dtype == np.min_scalar_type(nb)
    assert result.known_divisions
    assert result.npartitions == nb
    assert all(result.columns == ["band", name])
    assert truth.equals(resultc)



from functools import partial

import dask.array as da
import geopandas as gpd
import numba as nb
import numpy as np
import pandas as pd
import shapely
import xarray as xr

from raster_tools.dtypes import F32, I8, is_bool, is_scalar, is_str
from raster_tools.raster import Raster, get_raster
from raster_tools.utils import single_band_mappable
from raster_tools.vector import _geoms_to_raster_mask, get_vector


def _trim(x, slices):
    return x[tuple(slices)]


@nb.jit(nopython=True, nogil=True)
def _length_sum(out, lengths, idxs):
    n = len(lengths)
    for i in range(n):
        out[idxs[i]] += lengths[i]


def _compute_lengths(geom_df, targets):
    """
    Overlay the targets circles on the geometries in `geometry`. This produces
    a clip of `geometry` for each target. The output is the computed length of
    each clip result. The output lengths array is the same shape as the input
    targets and the elements correspond to the same location in `targets`.

    See geopandas.overlay for more details about the overlay operation.
    """
    targets_df = targets.to_frame("geometry")
    # Add column of index values so that clipped geoms can be mapped back to
    # the target that clipped them after the overlay operation. Column values
    # are preserved in overlay.
    targets_df["target_idx"] = np.arange(len(targets))
    overlay = geom_df.overlay(targets_df, keep_geom_type=True)
    overlay["len"] = overlay.length
    if "weights" in geom_df:
        overlay["len"] *= overlay.weights
    lengths = np.zeros(len(targets), dtype=F32)
    _length_sum(lengths, overlay.len.values, overlay.target_idx.values)
    return lengths


def _get_valid_slice(xc):
    # Build a slice that spans the valid data within the given coordinate array
    left = None
    i = 0
    while np.isnan(xc[i]):
        left = i + 1
        i += 1
    right = None
    i = -1
    while np.isnan(xc[i]):
        right = i
        i -= 1
    return slice(left, right)


def _len_coords(geom):
    return len(geom.coords)


def _len_interior_coords(geoms):
    return sum(map(_len_coords, geoms))


def _calculate_number_vertices(geoms):
    # Assume all geoms are valid
    has_exterior = geoms.geom_type == "Polygon"
    has_interior = ~geoms[has_exterior].interiors.isin([[]])
    nverts = (geoms.geom_type == "Point").sum()
    if has_exterior.any():
        nverts += (
            geoms[has_exterior].exterior.apply(_len_coords).astype(int).sum()
        )
        if has_interior.any():
            nverts += (
                geoms[has_exterior][has_interior]
                .interiors.apply(_len_interior_coords)
                .astype(int)
                .sum()
            )
    nverts += (
        geoms[geoms.geom_type.isin(set(["LineString", "LinearRing"]))]
        .apply(_len_coords)
        .astype(int)
        .sum()
    )
    return nverts


# Buffering a point with the following resolutions yields polygons with the
# corresponding area percentage of a circle:
# resolution | % area of circle
# -----------+-----------------
#      16    |      99.8
#       8    |      99.4
#       4    |      97.4
#       2    |      90.0
# -----------+-----------------
# The resolution is the number of segments along a quarter arc of a circle so
# the total number of coordinate pairs for a buffered point is 4x the
# resolution. The default resolution for buffering in geopandas and shapely is
# 16 but this is very expensive, in terms of time and memory, for very large
# geometries or large collections of geometries. 8 is also expensive. 4 seems
# to be a good compromise between performance and accuracy.
BUFFER_RES = 4
# The max desired average vertices per geometry in a geoseries. Values higher
# than this tend to cause large fluctuations in memory usage. When combined
# with dask's scheduling, the system has a high chance of running out of
# memory.
VERTS_PER_GEOM_CUTOFF = 15
# The max number of geometries to work on at a time. Large collections of
# geometries rapidly consume large amounts of memory when buffered and
# rasterized. Keeping the number processed at a time low allows dask to better
# handle the memory load.
GEOM_BATCH_SIZE = 700


def _get_indices_and_lengths_core(geoms_df, xc, yc, radius):
    # Rasterize to create mask of cells-of-interest and extract the locations
    indices = _geoms_to_raster_mask(
        xc, yc, geoms_df.geometry.buffer(radius, BUFFER_RES)
    ).nonzero()
    # Convert to points and buffer them out by the radius
    buffered_cells = gpd.GeoSeries.from_xy(
        xc[indices[1]], yc[indices[0]], crs=geoms_df.crs
    ).buffer(radius, BUFFER_RES)
    # Get length of all lines that fall within each buffered area.
    lengths = _compute_lengths(geoms_df, buffered_cells)
    return indices, lengths


def _get_indices_and_lengths(geoms_df, xc, yc, radius, weight_values=None):
    """
    Finds the indices of cells where geoms fall within the given radius and
    computes the lengths of those geoms for each one of the cells.

    If the number of geoms is too large or the geoms are too complex, the data
    is broken into batches. Batches are used because this process is memory
    intensive and can cause large swings in memory use. Dask can inadvertently
    spawn too many workers so that when enough memory swings occur at once,
    system memory is totally consumed and the whole program is killed. The
    smaller batches smooth out the memory consumption and prevent crashing.
    """
    # Explode to turn multi-geometries into single geometries for vertex
    # counting
    geoms_df = geoms_df.explode(ignore_index=True)
    nverts = _calculate_number_vertices(geoms_df.geometry)
    n = len(geoms_df)
    ratio = nverts / n
    if (ratio > VERTS_PER_GEOM_CUTOFF and n > 1) or n > GEOM_BATCH_SIZE:
        n_batches = max(1, int(ratio / VERTS_PER_GEOM_CUTOFF))
        # Clamp between 1 and GEOM_BATCH_SIZE
        batch_size = min(GEOM_BATCH_SIZE, max(1, n // n_batches))
        dfs = []
        # Process the batches
        for i in range(0, n, batch_size):
            batch_df = geoms_df.iloc[i : i + batch_size]
            idxs, lens = _get_indices_and_lengths_core(
                batch_df, xc, yc, radius
            )
            df = pd.DataFrame(
                {"idx0": idxs[0], "idx1": idxs[1], "length": lens}
            )
            dfs.append(df)
        # Merge results, summing co-located length values
        dfs = pd.concat(dfs, ignore_index=True)
        dfs = dfs.groupby(["idx0", "idx1"], as_index=False).sum()
        indices = tuple(dfs[["idx0", "idx1"]].to_numpy().T)
        lengths = dfs["length"].to_numpy()
    else:
        indices, lengths = _get_indices_and_lengths_core(
            geoms_df, xc, yc, radius
        )
    return indices, lengths


@single_band_mappable(no_input_chunk=True)
def _length_chunk(xc, yc, gdf, radius, field=None, block_info=None):
    xc = xc.ravel()
    yc = yc.ravel()
    # The overlap operation extends the raster into regions with no coords
    # given. The coords in these regions are set to nan. Find the nan coord
    # spans and trim them off.
    xslice = _get_valid_slice(xc)
    yslice = _get_valid_slice(yc)
    xc_valid = xc[xslice]
    yc_valid = yc[yslice]
    out_shape = block_info[None]["chunk-shape"][-2:]

    # Filter out any geoms that could be problematic
    # TODO: throw error if bad geoms detected?
    gdf = gdf[(~gdf.is_empty) & gdf.is_valid]
    # Clip out any features outside of this chunk's area of interest
    gdf = gdf.clip(
        shapely.geometry.box(
            xc_valid.min(),
            yc_valid.min(),
            xc_valid.max(),
            yc_valid.max(),
        )
        .buffer(radius)
        .bounds
    )

    if len(gdf):
        cols = ["geometry"]
        if field is not None:
            cols.append(field)
        gdf = gdf[cols]
        if field is not None:
            # Use field as weights
            gdf = gdf.rename({field: "weights"})
            if is_bool(gdf.weights.dtype):
                gdf = gdf.astype({"weights": I8})

        indices, lengths = _get_indices_and_lengths(
            gdf, xc_valid, yc_valid, radius
        )
        output = np.zeros(out_shape, dtype=F32)
        output[yslice, xslice][indices] = lengths
    else:
        output = np.zeros(out_shape, dtype=F32)
    return output


def length(features, like_rast, radius, weighting_field=None):
    """
    Calculate a raster where each cell is the net length of all features within
    a given radius.

    This function returns a raster where each cell contains the sum of the
    lengths of all features that fall within a radius of the cell's center. The
    features are clipped to the circular neighborhood before the length is
    computed. Optional weighting can be added with `weighting_field`.

    Parameters
    ----------
    features : Vector, str
        The line features to compute lengths from.
    like_rast : Raster, str
        A raster to use as a reference grid and CRS. The output raster will be
        on the same grid.
    radius : scalar
        The radius around each grid cell's center to use. Features that fall
        within the radius are clipped to the resulting circle and their
        resulting lengths are summed. Additional weighting of the sum can be
        done using `weighting_field`. This should be in the same units as
        `like_rast`'s CRS.
    weighting_field : str, optional
        If not ``None``, this selects a field in `features` to use as a
        weighting factor in the sum of lengths.

    Returns
    -------
    Raster
        The resulting raster where each cell contains the sum of all feature
        lengths in the circular neighborhood determined by `radius`.


    """
    features = get_vector(features)
    like_rast = get_raster(like_rast)

    if not is_scalar(radius):
        raise TypeError(f"radius must be a scalar. Got type: {type(radius)}.")
    if radius <= 0:
        raise ValueError(f"radius must be greater than zero: Got: {radius!r}.")
    if like_rast.crs is None:
        raise ValueError("like_rast must have a CRS set.")
    if weighting_field is not None:
        if not is_str(weighting_field):
            raise TypeError("weighting_field must be a string.")
        if weighting_field not in features.data:
            raise ValueError(
                "weighting_field must be a field name in features. "
                f"Got: {weighting_field!r}."
            )
        field_dtype = features.data[weighting_field].dtype
        if not (is_scalar(field_dtype) or is_bool(field_dtype)):
            raise TypeError(
                "The field specified by weighting_field must be a scalar type."
                f" Found: {field_dtype}."
            )

    features = features.to_crs(like_rast.crs)
    gdf = features.data
    out_chunks = ((1,), *like_rast.xdata.chunks[1:])

    # Cast to float in case coordinates are int. This allows for nan values in
    # overlap.
    xc, yc = [c.astype(float) for c in like_rast.get_chunked_coords()]
    xdepth, ydepth = np.ceil(radius / np.abs(like_rast.resolution)).astype(int)

    oxc = da.overlap.overlap(
        xc, depth={0: 0, 1: 0, 2: xdepth}, boundary=np.nan
    )
    oyc = da.overlap.overlap(
        yc, depth={0: 0, 1: ydepth, 2: 0}, boundary=np.nan
    )
    overlap_chunks = ((1,), oyc.chunks[1], oxc.chunks[2])
    rasters = []
    for part in gdf.partitions:
        data = da.map_blocks(
            _length_chunk,
            oxc,
            oyc,
            gdf=part,
            radius=radius,
            field=weighting_field,
            chunks=overlap_chunks,
            meta=np.array((), dtype=F32),
        )
        rasters.append(data[0])
    data = da.stack(rasters)
    # Trim off overlap regions
    trim_func = partial(
        _trim,
        slices=(slice(None), slice(ydepth, -ydepth), slice(xdepth, -xdepth)),
    )
    data = da.map_blocks(
        trim_func,
        data,
        chunks=((1,) * data.shape[0], *out_chunks[1:]),
        meta=np.array((), dtype=F32),
    )
    data = da.sum(data, axis=0, keepdims=True)
    xrs = xr.DataArray(
        data, dims=("band", "y", "x"), coords=([1], like_rast.y, like_rast.x)
    ).rio.write_crs(like_rast.crs)
    return Raster(xrs)



import os
from functools import wraps

import numpy as np
import xarray as xr


def validate_file(path):
    if os.path.isfile(path):
        return path
    raise FileNotFoundError(f"Could not find file: '{path}'")


def validate_path(path):
    if os.path.exists(path):
        return path
    raise FileNotFoundError(f"Path does not exist: '{path}'")


def is_strictly_increasing(x):
    x = np.atleast_1d(np.asarray(x).squeeze())
    assert x.ndim == 1

    diff = np.diff(x)
    return len(diff) > 0 and (diff > 0).all()


def is_strictly_decreasing(x):
    x = np.atleast_1d(np.asarray(x).squeeze())
    assert x.ndim == 1

    diff = np.diff(x)
    return len(diff) > 0 and (diff < 0).all()


def can_broadcast(*shapes, max_dim=3, min_dim=3):
    try:
        result = np.broadcast_shapes(*shapes)
        return len(result) >= min_dim and len(result) <= max_dim
    except ValueError:
        return False


def make_raster_ds(raster_dataarray, mask_dataarray):
    return xr.Dataset({"raster": raster_dataarray, "mask": mask_dataarray})


def merge_masks(masks):
    mask = None
    for m in masks:
        if mask is None:
            mask = m
        else:
            mask |= m
    return mask


def single_band_mappable(
    func_=None, *, no_input_chunk=False, pass_block_info=False
):
    """
    Decorator allowing functions that operate on 2D chunks to handle single
    band 3D chunks.

    This allows functions to be mapped to (1, y, x) chunks in a dask map_blocks
    operation.

    """

    if not no_input_chunk:

        def decorator(func):
            @wraps(func)
            def wrapper(chunk, *args, block_info=None, **kwargs):
                has_band_dim = chunk.ndim == 3
                if has_band_dim:
                    if chunk.shape[0] > 1:
                        raise TypeError(
                            f"Expexted 0 or 1 bands. Got {chunk.shape[0]}."
                        )
                    chunk = chunk[0]
                if not pass_block_info:
                    result = func(chunk, *args, **kwargs)
                else:
                    result = func(
                        chunk, *args, block_info=block_info, **kwargs
                    )
                if has_band_dim:
                    if result.ndim > 2:
                        raise TypeError(
                            "Expected result to have 2 dimensions. "
                            f"Got {result.ndim}."
                        )
                    result = np.expand_dims(result, axis=0)
                return result

            return wrapper

    else:

        def decorator(func):
            @wraps(func)
            def wrapper(*args, block_info=None, **kwargs):
                result = func(*args, block_info=block_info, **kwargs)
                if (
                    block_info is not None
                    and result.ndim < 3
                    and len(block_info[None]["chunk-shape"]) == 3
                ):
                    result = np.expand_dims(result, axis=0)
                return result

            return wrapper

    if func_ is None:
        return decorator
    return decorator(func_)



import numba as nb
import numpy as np
from numba.types import Tuple

from raster_tools.dtypes import F64, I64

__all__ = ["HEAP_TYPE", "init_heap_data", "pop", "push", "push_if_lower"]

JIT_KWARGS = {"nopython": True, "nogil": True}

HEAP_SPEC = [
    ("max_value", np.int64),
    ("levels", np.int8),
    ("min_levels", np.int8),
    ("count", np.int64),
]
HEAP_DT = np.dtype(HEAP_SPEC)
HEAP_TYPE = nb.from_dtype(HEAP_DT)

HEAP_STATE_SIG = Tuple(
    (
        nb.float64[:],
        nb.int64[:],
        nb.int64[:],
        HEAP_TYPE[:],
    )
)
(
    KEYS_TYPE_SIG,
    VALUES_TYPE_SIG,
    CROSSREFS_TYPE_SIG,
    HEAP_TYPE_SIG,
) = HEAP_STATE_SIG


@nb.jit(HEAP_STATE_SIG(nb.int64, nb.int64), **JIT_KWARGS)
def init_heap_data(capacity: I64, max_value: I64):
    if capacity < 1:
        raise ValueError("Capacity must be greater than 0")
    if max_value < 1:
        raise ValueError("Max value must be greater than 0")
    # Have to use single element array. Numba throws an error if you try to
    # return a record dtype. Using the array as a wrapper allows the record to
    # be passed around in these functions.
    heap = np.zeros(1, dtype=HEAP_DT)
    heap[0].max_value = max_value
    heap[0].levels = 0
    while 2 ** heap[0].levels < capacity:
        heap[0].levels += 1
    heap[0].min_levels = heap[0].levels
    heap[0].count = 0

    n = 2 ** heap[0].levels
    keys = np.full(2 * n, np.inf, dtype=F64)
    values = np.full(n, -1, dtype=I64)
    crossrefs = np.full(heap[0].max_value + 1, -1, dtype=I64)
    return keys, values, crossrefs, heap


@nb.jit(nb.void(KEYS_TYPE_SIG, HEAP_TYPE_SIG, nb.int64), **JIT_KWARGS)
def _sift(keys, heap, i):
    # Get first index in the pair. Pairs always start at uneven indices
    i -= i % 2 == 0
    # sift the minimum value back through the heap
    for _ in range(heap[0].levels, 1, -1):
        # Index into previous level
        iplvl = (i - 1) // 2
        keys[iplvl] = min(keys[i], keys[i + 1])
        i = iplvl - (iplvl % 2 == 0)


@nb.jit(nb.void(KEYS_TYPE_SIG, HEAP_TYPE_SIG), **JIT_KWARGS)
def _sift_all(keys, heap):
    # index at start of final level
    ilvl = (1 << heap[0].levels) - 1
    # Length of final level
    n = ilvl + 1
    for i in range(ilvl, ilvl + n, 2):
        _sift(keys, heap, i)


@nb.jit(
    Tuple((KEYS_TYPE_SIG, VALUES_TYPE_SIG))(
        KEYS_TYPE_SIG, VALUES_TYPE_SIG, HEAP_TYPE_SIG, nb.int8
    ),
    **JIT_KWARGS,
)
def _resize_levels(keys, values, heap, inc_dec):
    new_levels = heap[0].levels + inc_dec
    n = 1 << new_levels
    # We don't need to resize the crossrefs array
    new_keys = np.full(2 * n, np.inf, dtype=F64)
    new_values = np.full(n, -1, dtype=I64)
    if heap[0].count:
        inew = (1 << new_levels) - 1
        iold = (1 << heap[0].levels) - 1
        n = min(inew, iold) + 1
        new_keys[inew : inew + n] = keys[iold : iold + n]
        new_values[:n] = values[:n]
    keys = new_keys
    values = new_values
    heap[0].levels = new_levels
    _sift_all(keys, heap)
    return keys, values


@nb.jit(
    Tuple((KEYS_TYPE_SIG, VALUES_TYPE_SIG))(*HEAP_STATE_SIG, nb.int64),
    **JIT_KWARGS,
)
def _remove(keys, values, crossrefs, heap, i):
    lvl_start = (1 << heap[0].levels) - 1
    ilast = lvl_start + heap[0].count - 1
    i_val = i - lvl_start
    i_val_last = heap[0].count - 1

    # Get the last value
    value = values[i_val_last]
    # and update the crossref to point to its new location
    crossrefs[value] = i_val
    # Get the value to be removed
    value = values[i_val]
    # and delete its crossref
    crossrefs[value] = -1
    # Swap key to be removed with the last key
    keys[i] = keys[ilast]
    # Swap corresponding values
    values[i_val] = values[i_val_last]
    # Remove the last key since it is now invalid
    keys[ilast] = np.inf

    heap[0].count -= 1
    if (heap[0].levels > heap[0].min_levels) & (
        heap[0].count < (1 << (heap[0].levels - 2))
    ):
        keys, values = _resize_levels(keys, values, heap, -1)
    else:
        _sift(keys, heap, i)
        _sift(keys, heap, ilast)
    return keys, values


@nb.jit(
    HEAP_STATE_SIG(*HEAP_STATE_SIG, nb.float64, nb.int64),
    **JIT_KWARGS,
)
def _simple_push(keys, values, crossrefs, heap, key, value):
    lvl_size = 1 << heap[0].levels

    if heap[0].count >= lvl_size:
        keys, values = _resize_levels(keys, values, heap, 1)
        lvl_size = lvl_size << 1
    i = lvl_size - 1 + heap[0].count
    keys[i] = key
    values[heap[0].count] = value
    crossrefs[value] = heap[0].count
    heap[0].count += 1
    _sift(keys, heap, i)
    return keys, values, crossrefs, heap


@nb.jit(
    nb.types.Tuple((*HEAP_STATE_SIG, nb.int8))(
        *HEAP_STATE_SIG, nb.float64, nb.int64
    ),
    **JIT_KWARGS,
)
def push(keys, values, crossrefs, heap, key, value):
    if not (0 <= value <= heap[0].max_value):
        return keys, values, crossrefs, heap, -1

    lvl_size = 1 << heap[0].levels
    ii = crossrefs[value]
    if ii != -1:
        # Update the key value
        i = lvl_size - 1 + ii
        keys[i] = key
        _sift(keys, heap, i)
        return keys, values, crossrefs, heap, 0

    keys, values, crossrefs, heap = _simple_push(
        keys, values, crossrefs, heap, key, value
    )
    return keys, values, crossrefs, heap, 1


@nb.jit(
    nb.types.Tuple((*HEAP_STATE_SIG, nb.int8))(
        *HEAP_STATE_SIG, nb.float64, nb.int64
    ),
    **JIT_KWARGS,
)
def push_if_lower(keys, values, crossrefs, heap, key, value):
    if not (0 <= value <= heap[0].max_value):
        return keys, values, crossrefs, heap, -1

    icr = crossrefs[value]
    if icr != -1:
        i = (1 << heap[0].levels) - 1 + icr
        if keys[i] > key:
            keys[i] = key
            _sift(keys, heap, i)
            return keys, values, crossrefs, heap, 1
        return keys, values, crossrefs, heap, 0

    keys, values, crossrefs, heap = _simple_push(
        keys, values, crossrefs, heap, key, value
    )
    return keys, values, crossrefs, heap, 1


@nb.jit(
    nb.types.Tuple((*HEAP_STATE_SIG, nb.float64, nb.int64))(*HEAP_STATE_SIG),
    **JIT_KWARGS,
)
def pop(keys, values, crossrefs, heap):
    # The minimum key sits at position 1
    i = 1
    # Trace min key through the heap levels
    for lvl in range(1, heap[0].levels):
        if keys[i] <= keys[i + 1]:
            i = (i * 2) + 1
        else:
            i = ((i + 1) * 2) + 1
    # Find it in the last level
    if keys[i] > keys[i + 1]:
        i += 1
    # Get corresponding index into values
    ii = i - ((1 << heap[0].levels) - 1)
    popped_key = keys[i]
    popped_value = values[ii]
    if heap.count:
        keys, values = _remove(keys, values, crossrefs, heap, i)
    return keys, values, crossrefs, heap, popped_key, popped_value



import unittest

import numpy as np
from affine import Affine

from raster_tools import Raster, distance
from tests.utils import assert_valid_raster

# Example taken from ESRI docs
SOURCES = np.array(
    [
        [0, 1, 1, 0, 0, 0],
        [0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [2, 0, 0, 0, 0, 0],
    ]
)
SOURCES_IDXS = np.argwhere(SOURCES != 0)
COST_SURF = np.array(
    [
        [1, 3, 4, 4, 3, 2],
        [7, 3, 2, 6, 4, 6],
        [5, 8, 7, 5, 6, 6],
        [1, 4, 5, -1, 5, 1],
        [4, 7, 5, -1, 2, 6],
        [1, 2, 2, 1, 3, 4],
    ]
)
# Cost dist truths
CD_TRUTH_SCALE_1 = np.array(
    [
        [
            [2.0, 0.0, 0.0, 4.0, 7.5, 10.0],
            [6.0, 2.5, 0.0, 4.0, 9.0, 13.86396103],
            [8.0, 7.07106781, 4.5, 4.94974747, 10.44974747, 12.74264069],
            [5.0, 7.5, 10.5, -1.0, 10.62132034, 9.24264069],
            [2.5, 5.65685425, 6.44974747, -1.0, 7.12132034, 11.12132034],
            [0.0, 1.5, 3.5, 5.0, 7.0, 10.5],
        ]
    ]
)
CD_TRUTH_SCALE_5 = np.array(
    [
        [
            [10.0, 0.0, 0.0, 20.0, 37.5, 50.0],
            [30.0, 12.5, 0.0, 20.0, 45.0, 69.31980515],
            [40.0, 35.35533906, 22.5, 24.74873734, 52.24873734, 63.71320344],
            [25.0, 37.5, 52.5, -1.0, 53.10660172, 46.21320344],
            [12.5, 28.28427125, 32.24873734, -1.0, 35.60660172, 55.60660172],
            [0.0, 7.5, 17.5, 25.0, 35.0, 52.5],
        ]
    ]
)
# Traceback Truths
TR_TRUTH_SCALE_1 = np.array(
    [
        [
            [1, 0, 0, 5, 5, 5],
            [7, 1, 0, 5, 5, 6],
            [3, 8, 7, 6, 5, 3],
            [3, 5, 7, -1, 3, 4],
            [3, 4, 4, -1, 4, 5],
            [0, 5, 5, 5, 5, 5],
        ]
    ],
)
# Allocation truths
AL_TRUTH_SCALE_1 = np.array(
    [
        [
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [2, 1, 1, 1, 1, 2],
            [2, 2, 1, 0, 2, 2],
            [2, 2, 2, 0, 2, 2],
            [2, 2, 2, 2, 2, 2],
        ]
    ]
)
AL_TRUTH_IDXS_SCALE_1 = np.array(
    [
        [
            [0, 0, 1, 1, 1, 1],
            [0, 2, 2, 2, 2, 1],
            [3, 2, 2, 2, 2, 3],
            [3, 3, 2, -1, 3, 3],
            [3, 3, 3, -1, 3, 3],
            [3, 3, 3, 3, 3, 3],
        ]
    ]
)


class TestCostDist(unittest.TestCase):
    def setUp(self):
        self.cs = Raster(COST_SURF).set_null_value(-1)
        self.srcs = Raster(SOURCES).set_null_value(0)
        self.srcs_idx = SOURCES_IDXS

    def test_cost_distance_analysis(self):
        # ** Using srcs raster **
        cost_dist, traceback, allocation = distance.cost_distance_analysis(
            self.cs, self.srcs
        )
        assert_valid_raster(cost_dist)
        assert_valid_raster(traceback)
        assert_valid_raster(allocation)

        # Cost dist
        self.assertTrue(
            np.allclose(cost_dist.values, CD_TRUTH_SCALE_1, equal_nan=True)
        )
        self.assertTrue(cost_dist._masked)
        self.assertTrue(cost_dist.dtype == np.dtype(np.float64))
        self.assertTrue(cost_dist.null_value == -1)
        # traceback
        self.assertTrue(
            np.allclose(traceback.values, TR_TRUTH_SCALE_1, equal_nan=True)
        )
        self.assertTrue(traceback._masked)
        self.assertTrue(traceback.dtype == np.dtype(np.int8))
        self.assertTrue(traceback.null_value == -1)
        # Allocation
        self.assertTrue(
            np.allclose(allocation.values, AL_TRUTH_SCALE_1, equal_nan=True)
        )
        self.assertTrue(allocation._masked)
        self.assertTrue(allocation.dtype == np.dtype(np.int64))
        self.assertTrue(allocation.null_value == self.srcs.null_value)

        # ** Using srcs indices **
        cost_dist, traceback, allocation = distance.cost_distance_analysis(
            self.cs, self.srcs_idx
        )
        assert_valid_raster(cost_dist)
        assert_valid_raster(traceback)
        assert_valid_raster(allocation)

        # Cost dist
        self.assertTrue(
            np.allclose(cost_dist.values, CD_TRUTH_SCALE_1, equal_nan=True)
        )
        self.assertTrue(cost_dist._masked)
        self.assertTrue(cost_dist.dtype == np.dtype(np.float64))
        self.assertTrue(cost_dist.null_value == -1)
        # traceback
        self.assertTrue(
            np.allclose(traceback.values, TR_TRUTH_SCALE_1, equal_nan=True)
        )
        self.assertTrue(traceback._masked)
        self.assertTrue(traceback.dtype == np.dtype(np.int8))
        self.assertTrue(traceback.null_value == -1)
        # Allocation
        self.assertTrue(
            np.allclose(
                allocation.values, AL_TRUTH_IDXS_SCALE_1, equal_nan=True
            )
        )
        self.assertTrue(allocation._masked)
        self.assertTrue(allocation.dtype == np.dtype(np.int64))
        self.assertTrue(allocation.null_value == -1)

    def test_cost_distance_analysis_errors(self):
        with self.assertRaises(ValueError):
            # Must be single band
            distance.cost_distance_analysis(
                "tests/data/raster/multiband_small.tif", self.srcs
            )
        with self.assertRaises(ValueError):
            # Must have same shape
            distance.cost_distance_analysis(
                self.cs, "tests/data/raster/elevation_small.tif"
            )
        with self.assertRaises(TypeError):
            # source raster must be int
            distance.cost_distance_analysis(
                "tests/data/raster/elevation_small.tif",
                "tests/data/raster/elevation_small.tif",
            )
        with self.assertRaises(ValueError):
            # source raster must have null value
            distance.cost_distance_analysis(
                "tests/data/raster/elevation_small.tif",
                Raster("tests/data/raster/elevation_small.tif")
                .astype(int, False)
                .set_null_value(None),
            )
        with self.assertRaises(ValueError):
            # sources array must have shape (M, 2)
            distance.cost_distance_analysis(
                self.cs, np.zeros((5, 3), dtype=int)
            )
        with self.assertRaises(ValueError):
            # sources array must not have duplicates
            distance.cost_distance_analysis(
                self.cs, np.zeros((5, 2), dtype=int)
            )

    def test_cost_distance_analysis_scale(self):
        af = self.cs.affine
        coefs = [5, af.b, af.c, af.d, 5, af.f]
        trans = Affine(*coefs)
        cs = self.cs.copy()
        ds = self.cs._ds.rio.write_transform(trans)
        ds["x"] = ds.x * 5
        ds["y"] = ds.y * 5
        cs._ds = ds
        cost_dist, traceback, allocation = distance.cost_distance_analysis(
            cs, self.srcs
        )
        assert_valid_raster(cost_dist)
        assert_valid_raster(traceback)
        assert_valid_raster(allocation)

        self.assertTrue(
            np.allclose(cost_dist.values, CD_TRUTH_SCALE_5, equal_nan=True)
        )


def test_cost_distance_analysis_crs():
    rs = Raster("tests/data/raster/elevation_small.tif")
    srcs = np.array([[1, 1], [20, 30]])
    cd, tr, al = distance.cost_distance_analysis(rs, srcs)
    assert_valid_raster(cd)
    assert_valid_raster(tr)
    assert_valid_raster(al)
    assert cd.crs == rs.crs
    assert tr.crs == rs.crs
    assert al.crs == rs.crs



from collections.abc import Sequence

import dask.array as da
import numpy as np
import xarray as xr

from raster_tools.dtypes import is_int, is_scalar
from raster_tools.raster import Raster, get_raster
from raster_tools.utils import make_raster_ds

__all__ = [
    "constant_raster",
    "empty_like",
    "full_like",
    "ones_like",
    "random_raster",
    "zeros_like",
]

_VALID_RANDOM_DISTRIBUTIONS = frozenset(
    (
        "b",
        "binomial",
        "n",
        "normal",
        "p",
        "poisson",
        "u",
        "uniform",
        "w",
        "webull",
    )
)


def random_raster(
    raster_template, distribution="normal", bands=1, params=(1, 0.5)
):
    """Creates a Raster of random values based on the desired distribution.

    This function uses dask.array.random to generate data. The default is a
    normal distribution with mean of 1 and standard deviation of 0.5.

    Parameters
    ----------
    raster_template : Raster, str
        A template raster used to define rows, columns, crs, resolution, etc.
    distribution : str, optional
        Random distribution type. Default is `'normal'`. See `params` parameter
        below for passing additional distribution  parameters. Valid values
        are:

        'binomial'
            Binomial distribution. Uses two additional parameters: (n, p).
        'normal'
            Normal distribution. Uses two additional parameters: (mean, std).
            This is the default.
        'poisson'
            Poisson distribution. Uses one additional parameter.
        'uniform'
            Uniform distribution in the half-open interval [0, 1). Uses no
            additional.
        'weibull'
            Weibull distribution. Uses one additional parameter.

    bands : int, optional
        Number of bands needed desired. Default is 1.
    params : list of scalars, optional
        Additional parameters for generating the distribution. For example
        `distribution='normal'` and `params=(1, 0.5)` would result in a normal
        distribution with mean of 1 and standard deviation of 0.5. Default is
        ``(1, 0.5)``.

    Returns
    -------
    Raster
        The resulting raster of random values pulled from the distribution.

    """
    rst = get_raster(raster_template)

    if not is_int(bands):
        try:
            bands = int(bands)
        except Exception:
            raise TypeError(
                f"Could not coerce bands argument to an int: {repr(bands)}"
            )
    if bands < 1:
        raise ValueError("Number of bands must be greater than 0")
    if not isinstance(params, Sequence):
        try:
            params = list(params)
        except Exception:
            raise TypeError(
                f"Could not coerce params argument to a list: {repr(params)}"
            )
    else:
        params = list(params)

    shape = (bands,) + rst.shape[1:]
    chunks = ((1,) * bands,) + rst.data.chunks[1:]

    dist = distribution.lower()
    if dist not in _VALID_RANDOM_DISTRIBUTIONS:
        raise ValueError(f"Unknown distribution type: {repr(distribution)}")
    if dist not in ("u", "uniform") and len(params) == 0:
        raise ValueError(
            "Not enough additional parameters for the distribution"
        )

    if dist in ("n", "normal"):
        if len(params) < 2:
            raise ValueError(
                "Two few additional parameters for normal distribution"
            )
        ndata = da.random.normal(
            params[0], params[1], size=shape, chunks=chunks
        )
    elif dist in ("p", "poisson"):
        ndata = da.random.poisson(params[0], size=shape, chunks=chunks)
    elif dist in ("b", "binomial"):
        if len(params) < 2:
            raise ValueError(
                "Two few additional parameters for binomial distribution"
            )
        ndata = da.random.binomial(
            params[0], params[1], size=shape, chunks=chunks
        )
    elif dist in ("w", "weibull"):
        ndata = da.random.weibull(params[0], size=shape, chunks=chunks)
    elif dist in ("u", "uniform"):
        ndata = da.random.random(size=shape, chunks=chunks)
    # TODO: add more distributions

    cband = np.arange(bands) + 1
    xdata = xr.DataArray(
        ndata, coords=(cband, rst.y, rst.x), dims=("band", "y", "x")
    )
    if rst.crs is not None:
        xdata = xdata.rio.write_crs(rst.crs)
    xmask = xr.zeros_like(xdata, dtype=bool)
    return Raster(make_raster_ds(xdata, xmask), _fast_path=True)


def empty_like(raster_template, bands=1, dtype=None):
    """Create a Raster filled with uninitialized data like a template raster.

    Parameters
    ----------
    raster_template : Raster, str
        Template raster used to define rows, columns, crs, resolution, etc
    bands : int, optional
        Number of bands desired for output. Default is 1.
    dtype : data-type, optional
        Overrides the result dtype.

    Returns
    -------
    Raster
        The resulting raster of uninitialized data.

    """
    rst = get_raster(raster_template)

    if not is_int(bands):
        try:
            bands = int(bands)
        except Exception:
            raise TypeError(
                f"Could not coerce bands argument to an int: {repr(bands)}"
            )
    if bands < 1:
        raise ValueError("Number of bands must be greater than 0")
    if dtype is not None:
        try:
            dtype = np.dtype(dtype)
        except TypeError:
            raise ValueError(
                f"Could not understand dtype argument: {repr(dtype)}"
            )

    shape = (bands,) + rst.shape[1:]
    chunks = ((1,) * bands,) + rst.data.chunks[1:]
    ndata = da.empty(shape, chunks=chunks, dtype=dtype)
    xdata = xr.DataArray(
        ndata,
        coords=(np.arange(bands) + 1, rst.y, rst.x),
        dims=("band", "y", "x"),
    )
    if rst.crs is not None:
        xdata = xdata.rio.write_crs(rst.crs)
    xmask = xr.zeros_like(xdata, dtype=bool)
    return Raster(make_raster_ds(xdata, xmask), _fast_path=True)


def full_like(raster_template, value, bands=1, dtype=None):
    """Create a Raster filled with a contant value like a template raster.

    Parameters
    ----------
    raster_template : Raster, str
        Template raster used to define rows, columns, crs, resolution, etc
    value : scalar
        Value to fill result with.
    bands : int, optional
        Number of bands desired for output. Default is 1.
    dtype : data-type, optional
        Overrides the result dtype.

    Returns
    -------
    Raster
        The resulting raster of constant values.

    """
    rst = get_raster(raster_template)

    if not is_int(bands):
        try:
            bands = int(bands)
        except Exception:
            raise TypeError(
                f"Could not coerce bands argument to an int: {repr(bands)}"
            )
    if bands < 1:
        raise ValueError("Number of bands must be greater than 0")
    if not is_scalar(value):
        try:
            value = float(value)
        except Exception:
            raise TypeError(
                f"Could not coerce value argument to a scalar: {repr(value)}"
            )
    if dtype is not None:
        try:
            dtype = np.dtype(dtype)
        except TypeError:
            raise ValueError(
                f"Could not understand dtype argument: {repr(dtype)}"
            )

    shape = (bands,) + rst.shape[1:]
    chunks = ((1,) * bands,) + rst.data.chunks[1:]
    ndata = da.full(shape, value, chunks=chunks, dtype=dtype)
    xdata = xr.DataArray(
        ndata,
        coords=(np.arange(bands) + 1, rst.y, rst.x),
        dims=("band", "y", "x"),
    )
    if rst.crs is not None:
        xdata = xdata.rio.write_crs(rst.crs)
    xmask = xr.zeros_like(xdata, dtype=bool)
    return Raster(make_raster_ds(xdata, xmask), _fast_path=True)


def constant_raster(raster_template, value=1, bands=1):
    """Create a Raster filled with a contant value like a template raster.

    This is a convenience function that wraps :func:`full_like`.

    Parameters
    ----------
    raster_template : Raster, str
        Template raster used to define rows, columns, crs, resolution, etc
    value : scalar, optional
        Value to fill result with. Default is 1.
    bands : int, optional
        Number of bands desired for output. Default is 1.

    Returns
    -------
    Raster
        The resulting raster of constant values.

    """
    return full_like(raster_template, value, bands=bands)


def zeros_like(raster_template, bands=1, dtype=None):
    """Create a Raster filled with zeros like a template raster.

    Parameters
    ----------
    raster_template : Raster, str
        Template raster used to define rows, columns, crs, resolution, etc
    bands : int, optional
        Number of bands desired for output. Default is 1.
    dtype : data-type, optional
        Overrides the result dtype.

    Returns
    -------
    Raster
        The resulting raster of zreos.

    """
    return full_like(raster_template, 0, bands=bands, dtype=dtype)


def ones_like(raster_template, bands=1, dtype=None):
    """Create a Raster filled with ones like a template raster.

    Parameters
    ----------
    raster_template : Raster, str
        Template raster used to define rows, columns, crs, resolution, etc
    bands : int, optional
        Number of bands desired for output. Default is 1.
    dtype : data-type, optional
        Overrides the result dtype.

    Returns
    -------
    Raster
        The resulting raster of ones.

    """
    return full_like(raster_template, 1, bands=bands, dtype=dtype)



import raster_tools._compat  # noqa: F401
from raster_tools import clipping, distance, focal, line_stats, surface, zonal
from raster_tools._version import __version__  # noqa
from raster_tools.creation import (
    constant_raster,
    empty_like,
    full_like,
    ones_like,
    random_raster,
    zeros_like,
)
from raster_tools.general import band_concat, reclassify, remap_range
from raster_tools.raster import Raster, get_raster
from raster_tools.vector import (
    Vector,
    count_layer_features,
    list_layers,
    open_vectors,
)

__all__ = [
    "Raster",
    "Vector",
    "band_concat",
    "clipping",
    "constant_raster",
    "count_layer_features",
    "distance",
    "empty_like",
    "focal",
    "full_like",
    "get_raster",
    "line_stats",
    "list_layers",
    "ones_like",
    "open_vectors",
    "random_raster",
    "reclassify",
    "remap_range",
    "surface",
    "zeros_like",
    "zonal",
]



from unittest import TestCase

import dask
import numpy as np

from raster_tools import clipping
from raster_tools.raster import Raster, RasterNoDataError
from raster_tools.vector import open_vectors
from tests.utils import assert_valid_raster


class TestClipping(TestCase):
    def setUp(self):
        self.dem = Raster("tests/data/raster/elevation.tif")
        self.pods = open_vectors("tests/data/vector/pods.shp")
        self.v10 = self.pods[10]
        self.v10_bounds = dask.compute(self.v10.to_crs(self.dem.crs).bounds)[0]

    def test_core_clip_out_dtype(self):
        result = clipping._clip(self.pods, self.dem)
        assert_valid_raster(result)
        self.assertTrue(result.dtype == self.dem.dtype)
        self.assertTrue(result.eval().dtype == self.dem.dtype)

    def test_clip(self):
        res = clipping.clip(self.v10, self.dem)
        truth = Raster("tests/data/raster/clipping_clip_pods_10.tif")
        assert_valid_raster(res)
        self.assertTrue(np.allclose(res, truth))
        assert res.crs == self.dem.crs

        res = clipping.clip(self.v10, self.dem, bounds=self.v10_bounds)
        self.assertTrue(np.allclose(res, truth))
        assert res.crs == self.dem.crs

    def test_erase(self):
        res = clipping.erase(self.v10, self.dem)
        truth = Raster("tests/data/raster/clipping_erase_pods_10.tif")
        assert_valid_raster(res)
        self.assertTrue(np.allclose(res, truth))
        assert res.crs == self.dem.crs

        res = clipping.erase(self.v10, self.dem, bounds=self.v10_bounds)
        self.assertTrue(np.allclose(res, truth))
        assert res.crs == self.dem.crs

    def test_mask(self):
        res = clipping.mask(self.v10, self.dem)
        truth = Raster("tests/data/raster/clipping_mask_pods_10.tif")
        assert_valid_raster(res)
        self.assertTrue(np.allclose(res, truth))
        assert res.crs == self.dem.crs

        res = clipping.mask(self.v10, self.dem, invert=True)
        truth = Raster("tests/data/raster/clipping_mask_inverted_pods_10.tif")
        assert_valid_raster(res)
        self.assertTrue(np.allclose(res, truth))
        assert res.crs == self.dem.crs

    def test_envelope(self):
        res = clipping.envelope(self.v10, self.dem)
        truth = Raster("tests/data/raster/clipping_envelope_pods_10.tif")
        assert_valid_raster(res)
        self.assertTrue(np.allclose(res, truth))
        assert res.crs == self.dem.crs

    def test_errors(self):
        with self.assertRaises(ValueError):
            rs = Raster(np.ones((4, 4)))
            clipping.clip(self.v10, rs)

        with self.assertRaises(ValueError):
            clipping._clip(self.v10, self.dem, bounds=(0, 3))

        with self.assertRaises(ValueError):
            clipping._clip(self.v10, self.dem, invert=True, envelope=True)

        with self.assertRaises(RuntimeError):
            clipping._clip(self.v10, self.dem, bounds=self.v10.bounds)

    def test_clip_box(self):
        self.dem = Raster("tests/data/raster/elevation.tif")
        rs_clipped = Raster("tests/data/raster/elevation_small.tif")
        bounds = [
            rs_clipped.xdata.x.min().item(),
            rs_clipped.xdata.y.min().item(),
            rs_clipped.xdata.x.max().item(),
            rs_clipped.xdata.y.max().item(),
        ]
        test = clipping.clip_box(self.dem, bounds)
        self.assertTrue(test.shape == rs_clipped.shape)
        self.assertTrue(np.allclose(test.values, rs_clipped.values))

        # Test that the mask is also clipped
        x = np.arange(25).reshape((1, 5, 5))
        x[x < 12] = 0
        rs = Raster(x).set_null_value(0).set_crs("epsg:3857")
        self.assertTrue(np.allclose(x == 0, rs.mask))
        rs_clipped = clipping.clip_box(rs, (1, 1, 4, 4))
        mask_truth = np.array([[[1, 1, 1], [1, 0, 0], [0, 0, 0]]], dtype=bool)
        self.assertTrue(np.allclose(rs_clipped.mask, mask_truth))

    def test_clip_out_of_bounds(self):
        with self.assertRaises(RasterNoDataError):
            clipping.clip_box(self.dem, (9e6, 9e6, 10e6, 10e6))
